# -*- coding: utf-8 -*-
#
#    TypeAtlas Font Grid
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#


"""Character map widget. Can be re-arranged into sample-viewing widget.

Includes tools for rendering individual characters (e.g. Emojis not
supported by the OS), and including an Emoji icon engine.
"""

from __future__ import division
from typeatlas.compat import QtCore, QtGui, QtWidgets, QtModelProxies
from typeatlas.compat import Qt, Slot, Signal, setResizeMode
try:
    from typeatlas.compat import QtSvg
except ImportError:
    QtSvg = None

import typeatlas
from typeatlas.util import OrderedSet, bytedata_to_int, debugmsg, warnmsgf
from typeatlas.util import generic_type
from typeatlas.uitools import Toolbox, qFontToCss, modelIterateChildrenRect
from typeatlas.uitools import getIconHtml, hasDarkBackground, TransposedModel
from typeatlas.uitools import RotatedHeaderView
from typeatlas.uitools import QuickSearch
from typeatlas.guicommon import FontRenderingChoice, CharInfoWidget, getIcon
from typeatlas.guicommon import CustomComboBox
from typeatlas.options import Options
from typeatlas.langutil import LanguageDatabase, textlang
from typeatlas.charinfo import CharacterDatabase
from typeatlas import charinfo, blockmath, osinfo, opentype, datastore
from itertools import chain
from functools import partial
from collections import Counter
from collections.abc import Mapping, Iterator
from html import escape as htesc
import re
import traceback
import functools
import numbers

try:
    from fontTools import ttLib
except ImportError:
    ttLib = None

from typeatlas.langutil import _, N_, H_

MutableSequenceOf = generic_type('MutableSequence')
SequenceOf = generic_type('Sequence')
MappingOf = generic_type('Mapping')
IterableOf = generic_type('Iterable')
SetOf = generic_type('Set')
TupleOf = generic_type('Tuple')
Optional = generic_type('Optional')
Union = generic_type('Union')


ID_THRESH = 22


PREVIEW_MAX_CODEPOINT = 256
BIG_FONT_THRESHOLD = 20000


CharCodeRole = Qt.UserRole
CellHeaderRole = Qt.UserRole + 1
SampleRendererRole = Qt.UserRole + 2
CharInfoTextRole = Qt.UserRole + 3
ModelFontInfoRole = Qt.UserRole + 4


class SvgFontSample(object):

    """Display a SVG font sample. Use fromSvgs to construct."""

    def __init__(self, sampleRenderers: 'SequenceOf[QtSvg.QSvgRenderer]',
                       antialias: bool=True):
        if len(sampleRenderers) != 1:
            raise NotImplementedError
        self.antialias = antialias
        self.renderers = sampleRenderers

    def forFont(self, font: QtGui.QFont) -> 'SvgFontSample':
        """Return the renderer for another font (e.g. when changing font size)."""
        return self

    @classmethod
    def fromSvgs(cls, svgs: SequenceOf[str],
                      antialias: bool=True) -> 'SvgFontSample':
        """Create a SVG font sample renderer from a list of SVG strings."""
        if len(svgs) != 1:
            raise NotImplementedError
        renderers = []
        for svg in svgs:
            svg = re.sub('viewBox=([\'"])([^\'"]+)(\\1)',  '', svg, 1)
            renderers.append(QtSvg.QSvgRenderer(QtCore.QByteArray(svg.encode('utf8'))))
        return cls(renderers, antialias)

    def render(self, painter: QtGui.QPainter, rect: QtCore.QRect,
                     palette: QtGui.QPalette=None, darkBackground: bool=None):
        """Render the sample using the provided painter, rectangle and palette.

        You can provide a darkBackground=True/False hint.
        """

        renderer = self.renderers[0]

        sampleRatio = rect.width() / rect.height()
        viewBox = renderer.viewBoxF()
        viewBoxRatio = viewBox.width() / viewBox.height()

        if sampleRatio > viewBoxRatio:
            newWidth = viewBox.width() * (sampleRatio / viewBoxRatio)
            extraWidth = newWidth - viewBox.width()
            viewBox.adjust(-extraWidth / 2, 0, extraWidth / 2, 0)
            renderer.setViewBox(viewBox)
        elif sampleRatio < viewBoxRatio:
            newHeight = viewBox.height() * (viewBoxRatio / sampleRatio)
            extraHeight = newHeight - viewBox.height()
            viewBox.adjust(0, -extraHeight / 2, 0, extraHeight / 2)
            renderer.setViewBox(viewBox)

        renderer.render(painter, QtCore.QRectF(rect))


class ColorLayerFontSample(object):


    """Display a CPAL/COLR layered color font sample.
    Use fromTables to construct."""

    def __init__(self, rawFont: 'QtGui.QRawFont',
                       layersSequence: SequenceOf[SequenceOf[TupleOf[int, int]]],
                       fontPalettes: SequenceOf[SequenceOf[object]]):
        if len(layersSequence) != 1:
            raise NotImplementedError
        self.rawFont = rawFont
        self.layersSequence = layersSequence

        self.fontPalettes = fontPalettes

        if not fontPalettes:
            self.darkFontPalette = self.lightFontPalette = ()
        elif len(fontPalettes) >= 2:
            self.lightFontPalette = fontPalettes[0]
            self.darkFontPalette = fontPalettes[1]
        else:
            self.darkFontPalette = self.lightFontPalette = fontPalettes[0]

    def forFont(self, font: QtGui.QFont) -> 'ColorLayerFontSample':
        """Return the renderer for another font (e.g. when changing font size)."""
        return type(self)(QtGui.QRawFont.fromFont(font),
                          self.layersSequence,
                          self.fontPalettes)

    @classmethod
    def fromTables(cls, rawFont: 'QtGui.QRawFont',
                        ttFont: 'ttLib.TTFont',
                        colr: 'ttLib.tables.C_O_L_R_.table_C_O_L_R_',
                        cpal: 'ttLib.tables.C_P_A_L_.table_C_P_A_L_',
                        glyphs: SequenceOf[str]):
        """Construct the renderer.

        You need to provide the TTFont, the COLR and CPAL tables of the font,
        and a list of glyph IDs to use.

        Currently only one glyph is supported, i.e. only character map, not
        sampling."""
        if len(glyphs) != 1:
            raise NotImplementedError

        layersSequence = []

        for glyph in glyphs:
            layers = colr.ColorLayers[glyph]
            layersSequence.append([(layer.colorID,
                                    ttFont.getGlyphID(layer.name))
                                   for layer in layers])
        return cls(rawFont, layersSequence, cpal.palettes)

    def render(self, painter: QtGui.QPainter, rect: QtCore.QRect,
                     palette: QtGui.QPalette=None, darkBackground: bool=None):
        """Render the sample using the provided painter, rectangle and palette.

        You can provide a darkBackground=True/False hint.
        """

        if palette is None:
            palette = QtWidgets.QApplication.palette()
        if darkBackground is None:
            darkBackground = hasDarkBackground(palette)
        if darkBackground:
            fontPalette = self.darkFontPalette
        else:
            fontPalette = self.lightFontPalette

        painter.save()
        ascent = self.rawFont.ascent()
        descent = self.rawFont.descent()
        fontHeight = ascent + descent
        rectHeight = rect.height()
        raiseWith = descent * rectHeight / fontHeight

        basePoint = QtCore.QPointF(rect.bottomLeft())
        basePoint += QtCore.QPointF(0, -raiseWith)

        for colorID, glyph in self.layersSequence[0]:
            if colorID == opentype.COLR_DEFAULT_TEXT_FOREGROUND:
                color = palette.color(palette.Text)

            else:
                try:
                    fontColor = fontPalette[colorID]
                except LookupError:
                    color = palette.color(palette.Text)
                else:
                    color = QtGui.QColor(fontColor.red, fontColor.green,
                                         fontColor.blue, fontColor.alpha)

            painter.setPen(color)
            glyphRun = QtGui.QGlyphRun()
            glyphRun.setRawFont(self.rawFont)
            glyphRun.setGlyphIndexes([glyph])
            glyphRun.setPositions([QtCore.QPointF(0, 0)])
            painter.setBrush(color)
            painter.drawGlyphRun(basePoint, glyphRun)

            #painterPath = self.rawFont.pathForGlyph(glyph)
            ##painterPath.translated(rect.topLeft())
            #painter.drawPath(painterPath)
        painter.restore()


SampleRendererType = Union[SvgFontSample, ColorLayerFontSample]


class CharacterIconEngine(QtGui.QIconEngine):

    """An icon engine for Emoji icon utilising a font that uses the
    sample renderers from above. It is created using an Emoji character,
    and optional font."""

    _renderModes = None

    def __init__(self, char: str, font: QtGui.QFont=None,
                       fontInfo: 'typeatlas.fontlist.FontLike'=None):
        super().__init__()
        if font is None:
            if fontInfo is None:
                font = QtGui.QFont()
            else:
                font = QtGui.QFontDatabase().font(fontInfo.family,
                                                  fontInfo.style, 32)
        self.char = char
        self.font = font
        self.fontInfo = fontInfo
        self._sampleRenderer = None
        self._sampleRendererDisabled = False

        #self._options = options

    @classmethod
    def _getRenderModes(cls) -> SetOf[str]:
        """Return the render modes supported by this class.
        This may support COLR and SVG modes."""

        if cls._renderModes is not None:
            return cls._renderModes

        options = Options.getInstance()

        if options.layeredFontRenderer == 'os' or not hasattr(QtGui, 'QGlyphRun'):
            renderColr = False
        elif options.layeredFontRenderer == 'internal':
            renderColr = True
        else:
            renderColr = not osinfo.layered_fonts_supported()

        if options.svgFontRenderer == 'os' or QtSvg is None:
            renderSvg = False
        elif options.svgFontRenderer == 'internal':
            renderSvg = True
        else:
            renderSvg = not osinfo.svg_fonts_supported()

        cls._renderModes = renderModes = set()
        if renderSvg:
            renderModes.add('SVG')
        if renderColr:
            renderModes.add('COLR')

        return renderModes

    def getSampleRenderer(self) -> Optional[SampleRendererType]:
        """Get the sample renderer for engine's character (text sample) and font.
        This does not throw exceptions, only displays them. We don't need to crash
        because we failed to load an icon font."""
        if self._sampleRendererDisabled:
            return None
        if self._sampleRenderer is not None:
            return self._sampleRenderer
        try:
            return self._sampleRendererWithExceptions()
        except Exception:
            warnmsgf("Internal rendering failed for %r", self.fontInfo)
            traceback.print_exc()
            self._sampleRendererDisabled = True

    def _sampleRendererWithExceptions(self) -> Optional[SampleRendererType]:
        """Get the sample renderer, throwing exceptions."""
        fontInfo = self.fontInfo

        if fontInfo is None:
            return None

        renderModes = self._getRenderModes()
        if not renderModes:
            return None

        # Keeping a reference to fontExtended, so other icons don't parse the
        # font again
        self.fontExtended = fontExtended = fontInfo.extended()

        if fontExtended.ttfont and fontExtended.cmap:
            char = self.char
            font = self.font

            charCode = ord(char)

            antialias = not (font.styleStrategy() & font.NoAntialias)
            ttFont = fontExtended.ttfont
            fontCmap = fontExtended.cmap

            try:
                svg = ttFont['SVG '] if 'SVG' in renderModes else None
            except KeyError:
                svg = None

            if svg and hasattr(svg, 'docList') and QtSvg is not None:
                glyphName = fontCmap.get(charCode)
                if glyphName:
                    glyphID = ttFont.getGlyphID(glyphName)
                    for svg, startGlyph, endGlyph in svg.docList:
                        if startGlyph <= glyphID <= endGlyph:
                            self._sampleRenderer = SvgFontSample.fromSvgs(
                                                            [svg], antialias)
                            return self._sampleRenderer

            if 'COLR' not in renderModes:
                colr = cpal = None
            else:
                try:
                    colr = ttFont['COLR']
                    cpal = ttFont['CPAL']
                except KeyError:
                    colr = cpal = None


            if colr is not None and hasattr(QtGui, 'QRawFont'):
                rawFont = QtGui.QRawFont.fromFont(font)

                glyph = fontCmap.get(charCode)
                if glyph is not None and glyph in colr.ColorLayers:

                    self._sampleRenderer = ColorLayerFontSample.fromTables(
                            rawFont, ttFont, colr, cpal, [glyph])
                    return self._sampleRenderer

    def icon(self) -> QtGui.QIcon:
        """Return icon backed by this engine."""
        return QtGui.QIcon(self)

    def paint(self, painter, rect, mode, state):

        #painter.fillRect(rect, QtGui.QColor("transparent"))

        sample = self.char
        style = QtWidgets.QApplication.style()
        palette = QtWidgets.QApplication.palette()

        font = QtGui.QFont(self.font)
        font.setPointSize(32)
        metrics = QtGui.QFontMetrics(font)

        boxWidth = rect.width()
        boxHeight = rect.height()

        sampleRect = metrics.boundingRect(sample)
        txtWidth = sampleRect.width()
        txtHeight = sampleRect.height()

        if (not boxWidth or not boxHeight or
            txtWidth / boxWidth <= txtHeight / boxHeight):
                font.setPixelSize(int(0.9 * boxHeight))
        else:
                font.setPixelSize(int(0.9 * txtHeight * boxWidth / txtWidth))

        metrics = QtGui.QFontMetrics(font)

        enabled = mode != QtGui.QIcon.Disabled

        if mode == QtGui.QIcon.Selected:
            samplePaintRole = palette.HighlightedText
        else:
            samplePaintRole = palette.Text

        sampleRenderer = self.getSampleRenderer()

        # Draw
        if mode == QtGui.QIcon.Selected:
            painter.fillRect(rect, palette.highlight())
        painter.setPen(palette.color(palette.Mid))

        painter.setFont(font)
        if sampleRenderer:
            sampleRenderer = sampleRenderer.forFont(font)
            sampleTextRect = style.itemTextRect(metrics, rect,
                                                Qt.AlignCenter, enabled,
                                                sample)
            sampleRenderer.render(painter, sampleTextRect, palette)
        else:
            style.drawItemText(painter,
                               style.itemTextRect(metrics, rect,
                                                  Qt.AlignCenter, enabled,
                                                  sample),
                               Qt.AlignCenter, palette, enabled, sample,
                               samplePaintRole)

    def pixmap(self, size, mode, state):
        pixmap = QtGui.QPixmap(size)
        pixmap.fill(QtGui.QColor(0, 0, 0, 0))
        painter = QtGui.QPainter()
        painter.begin(pixmap)
        self.paint(painter, QtCore.QRect(QtCore.QPoint(0, 0), size),
                   mode, state)
        painter.end()
        return pixmap


class CharSymbolInfo(object):

    """Info for a given character symbol in a given encoding."""

    def __init__(self, codepoint: int, bytedata: bytes=None,
                       charvalue: int=None,
                       display: str=None,
                       encoding: 'typeatlas.charsets.EncodingLike'=None):

        if encoding is not None:
            charsize = encoding.charsize
        else:
            charsize = None

        if charvalue is None:
            if bytedata is not None:
                if charsize == 1:
                    charvalue = bytedata[0]
                else:
                    charvalue = bytedata_to_int(bytedata)

        elif bytedata is None and charsize == 1:
            bytedata = bytes([charvalue])

        if display is None and bytedata is not None:
            if charsize == 1:
                display = "%X" % (charvalue, )
            else:
                display = ' '.join('%X' % (byte, ) for byte in bytedata)

        self.codepoint = codepoint
        self.charvalue = charvalue
        self.bytedata = bytedata
        self.display = display
        self.encoding = encoding


class SymbolModelFontInfo(object):

    """A font info object representing a font and its characters for the
    font grid character map model. A character map can display one or
    multiple fonts; this contains all font-specific attributes that will
    be repeated for each font. Used by setFont() and setMultiFont() of
    the FontSymbolModel to store information.

    This includes the Qt font, the fontlist.Font object, the characters in
    the font as blocks or a character iterable.

    You can provide charsetCodeInfo mapping of codepoints to CharSymbolInfo
    model. It can also be passed as characters= (as it is valid iterable of
    characters), and it will be detected properly.
    """

    def __init__(self, font: QtGui.QFont=None,
                       fontInfo: 'typeatlas.fontlist.FontLike'=None,
                       characters: IterableOf[int]=None,
                       charBlocks: IterableOf[blockmath.BlockLike]=None,
                       options: Options=None,
                       charsetCodeInfo: MappingOf[int, CharSymbolInfo]=None):
        if options is None:
            options = Options.getInstance()

        if options.layeredFontRenderer == 'os' or not hasattr(QtGui, 'QGlyphRun'):
            renderColr = False
        elif options.layeredFontRenderer == 'internal':
            renderColr = True
        else:
            renderColr = not osinfo.layered_fonts_supported()

        if options.svgFontRenderer == 'os' or QtSvg is None:
            renderSvg = False
        elif options.svgFontRenderer == 'internal':
            renderSvg = True
        else:
            renderSvg = not osinfo.svg_fonts_supported()

        if isinstance(characters, Mapping):
            if charsetCodeInfo is None:
                charsetCodeInfo = characters
            characters = list(characters)

        self.sampleRenderers = {}
        self._fontCharBlocks = charBlocks
        self._characters = characters

        self.charsetCodeInfo = charsetCodeInfo
        self.ownCharacters = (characters is None and charBlocks is None and
                              font is not None)

        self.encoding = None
        if self.charsetCodeInfo:
            for symbolInfo in self.charsetCodeInfo.values():
                self.encoding = symbolInfo.encoding
                break

        if font is None:
            self.font = None
            self.fontInfo = None
            self.fontExtended = None
            self.fontCmap = None
            return

        fontExtended = None

        rawFont = None
        fontCmap = None

        if fontInfo is not None:
            fontExtended = fontInfo.extended()
            if fontExtended.ttfont and fontExtended.cmap:
                antialias = not (font.styleStrategy() & font.NoAntialias)
                ttFont = fontExtended.ttfont
                fontCmap = fontExtended.cmap

                try:
                    svg = ttFont['SVG '] if renderSvg else None
                except KeyError:
                    svg = None

                if svg and hasattr(svg, 'docList') and QtSvg is not None:
                    charByGlyph = {ttFont.getGlyphID(glyphName): char
                                   for char, glyphName in fontCmap.items()}
                    for svg, startGlyph, endGlyph in svg.docList:
                        sampleRenderer = SvgFontSample.fromSvgs([svg],
                                                                antialias)
                        for glyph in range(startGlyph, endGlyph + 1):
                            char = charByGlyph.get(glyph)
                            if char is not None:
                                self.sampleRenderers[char] = sampleRenderer

                if not renderColr:
                    colr = cpal = None
                else:
                    try:
                        colr = ttFont['COLR']
                        cpal = ttFont['CPAL']
                    except KeyError:
                        colr = cpal = None

                if (rawFont is None and colr is not None and
                    hasattr(QtGui, 'QRawFont')):
                        rawFont = QtGui.QRawFont.fromFont(font)

                if colr is not None and rawFont is not None:
                    charByGlyph = {glyphName: char
                                   for char, glyphName in fontCmap.items()}
                    for glyph in colr.ColorLayers:
                        char = charByGlyph.get(glyph)
                        if char is None:
                            continue

                        sampleRenderer = ColorLayerFontSample.fromTables(
                                rawFont, ttFont, colr, cpal, [glyph])

                        if sampleRenderer is not None:
                            self.sampleRenderers[char] = sampleRenderer

        self.font = font
        self.fontInfo = fontInfo
        self.fontExtended = fontExtended
        self.fontCmap = fontCmap
        #self._characters = None
        #self._fontCharBlocks = None

    def characters(self) -> SetOf[int]:
        """Get the characters of the font as a list of integers."""
        if self._characters is None:
            self._fillCharacters()
        return self._characters

    def fontCharBlocks(self) -> SequenceOf[blockmath.BlockLike]:
        """Get the characters of the font as a list of character blocks."""
        if self._fontCharBlocks is None:
            self._fillCharacters()
        return self._fontCharBlocks

    def _fillCharacters(self):
        """Compute the characters of the font."""
        if self._characters is not None:
            self._fontCharBlocks = blockmath.toblocks(self._characters)
            return

        if self._fontCharBlocks is not None:
            self._characters = OrderedSet(blockmath.iterblocks(self._fontCharBlocks))
            return

        font = self.font
        fontInfo = self.fontInfo
        fontCmap = self.fontCmap

        if font is None and fontInfo is None and fontCmap is None:
            self._characters = frozenset()
            self._fontCharBlocks = []
            return

        if fontInfo is not None:
            charset = self.fontInfo.get_charset()
            if charset is not None:
                self._characters = charset
                self._fontCharBlocks = list(blockmath.toblocks(self._characters))
                return

        # TODO: Old code below, duplicated with above, some checks
        #       are senseless.

        # FIXME: Decide what interface to use here.
        if fontInfo is None or getattr(fontInfo, 'charblocks', None) is None:

            if fontCmap is not None:
                chars = sorted(fontCmap)
            elif hasattr(QtGui, 'QRawFont'):
                rawFont = QtGui.QRawFont.fromFont(font)
                allChars = charinfo.all_character_codes()
                chars = filter(rawFont.supportsCharacter, allChars)
            else:
                chars = charinfo.all_character_codes()

            charblocks = list(blockmath.toblocks_inorder(chars))
        else:
            charblocks = fontInfo.charblocks

        self._fontCharBlocks = charblocks
        characters = OrderedSet(blockmath.iterblocks(charblocks))

        #characters = []
        #for char in charinfo.all_character_codes():
        #    if rawFont.supportsCharacter(char):
        #         characters.append(char)
        self._characters = characters


class FontSymbolModel(QtCore.QAbstractItemModel):

    """A Qt item model containing font symbols for display in a character
    map using FontGrid().

    All extra attributes are forwarded to the SymbolModelFontInfo of the
    main font.

    It supports the following additional roles:
        CharCodeRole        The unicode code points.
        CellHeaderRole      The hex of the unicode point. The hex of the code
                            in the character set. The morse code for the letter.
                            To be shown in the cell's header.
        SampleRendererRole  The role returning the sample renderer for
                            the characters in fonts that need special rendering,
                            e.g. SVG or layered fonts (SampleRendererType)
        CharInfoTextRole    Unused
        ModelFontInfoRole   The SymbolModelFontInfo object for the current
                            column.

    """

    def __init__(self, font: QtGui.QFont=None,
                       renderingChoice: FontRenderingChoice=None,
                       parent: QtCore.QObject=None,
                       charDb: CharacterDatabase=None,
                       langDb: LanguageDatabase=None):

        super(FontSymbolModel, self).__init__(parent)

        if renderingChoice is None:
            renderingChoice = FontRenderingChoice()

        if charDb is None:
            charDb = CharacterDatabase.getInstance()
        if langDb is None:
            langDb = LanguageDatabase.getInstance()

        self.ownCharacters = False
        self.fontCharBlocks = []
        self.characters = []
        self.selected = []
        self.charsetCodeInfo = None

        self.renderingChoice = renderingChoice
        self._setFont(font)

        self.gridView = None
        self.charBox = None
        self.charDb = charDb
        self.langDb = langDb

        self.singleFamily = True
        self.singleStyle = True

    def __getattr__(self, attr):
        return getattr(self.modelFontInfo, attr)

    def _setFont(self, font: QtGui.QFont=None, *args, **kwargs):
        """Set the font. All arguments are passed to SymbolModelFontInfo()."""
        self.selected = []
        self.singleFamily = True
        self.singleStyle = True

        options = self.renderingChoice.options

        if font is not None and self.gridView is not None:
            pixelSize = font.pixelSize()
            if pixelSize >= ID_THRESH:
                factor = 1.3
            else:
                factor = 1.1
            self.gridView.setGridSize(QtCore.QSize(pixelSize * factor,
                                                   pixelSize * factor))

        self.modelFontInfo = SymbolModelFontInfo(font, *args, **kwargs,
                                                 options=options)
        self._setCharacters()

    def _setMultiFont(self, current: SymbolModelFontInfo,
                            selected: IterableOf[SymbolModelFontInfo]=()):
        """Set multiple fonts, one current and multiple selected, using
        explicit SymbolModelFontInfo instances."""
        options = self.renderingChoice.options
        self.modelFontInfo = current
        self.selected = list(selected)[:options.maxFontGridFonts]

        allFonts = [current] + self.selected

        familyNames = Counter(mfi.fontInfo.family for mfi in allFonts)
        self.singleFamily = len(familyNames) <= 1
        self.singleStyle = max(familyNames.values(), default=0) <= 1

        if self.gridView is not None and any(mfi.font is not None
                                             for mfi in allFonts):

            gridSize = max(mfi.font.pixelSize() *
                           (1.3 if mfi.font.pixelSize() >= ID_THRESH
                                else 1.1)
                           for mfi in allFonts
                           if mfi.font is not None)

            self.gridView.setGridSize(QtCore.QSize(gridSize, gridSize))

        self._setCharacters()

    def _setCharacters(self, characters: IterableOf[int]=None,
                             charBlocks: IterableOf[blockmath.BlockLike]=None,
                             charsetCodeInfo: MappingOf[int, CharSymbolInfo]=None):
        """Set the characters to display. If no characters are passed (the default),
        use the SymbolModelFontInfo objects to compute an intersection of
        characters."""
        if charBlocks is not None:
            self.fontCharBlocks = charBlocks
            self.characters = list(blockmath.iterblocks(charBlocks))
            self.ownCharacters = False
            self.charsetCodeInfo = charsetCodeInfo
            return

        if characters is not None:
            self.characters = list(characters)
            self.fontCharBlocks = blockmath.toblocks(self.characters)
            self.ownCharacters = False
            self.charsetCodeInfo = charsetCodeInfo
            return

        if self.modelFontInfo is None:
            return

        if self.selected:
            characters = self.modelFontInfo.characters()
            ownCharacters = self.modelFontInfo.ownCharacters
            self.charsetCodeInfo = self.modelFontInfo.charsetCodeInfo

            for mfi in self.selected:
                characters = characters & mfi.characters()
                if mfi.ownCharacters:
                    ownCharacters = True
                #self.charsetCodeInfo = mfi.charsetCodeInfo

            self.characters = list(characters)
            self.fontCharBlocks = blockmath.toblocks(self.characters)
            self.ownCharacters = self.ownCharacters

        else:
            self.characters = list(self.modelFontInfo.characters())
            self.fontCharBlocks = self.modelFontInfo.fontCharBlocks()
            self.ownCharacters = self.modelFontInfo.ownCharacters
            self.charsetCodeInfo = self.modelFontInfo.charsetCodeInfo

    def setFont(self, font: QtGui.QFont=None,
                      fontInfo: 'typeatlas.fontlist.FontLike'=None,
                      nochange: bool=False, **kwargs):
        """Set the font. See SymbolModelFontInfo for documentation of
        the arguments.

        Passing nochange=True means the model is not reset, just the
        display font is changed.
        """
        if nochange:
            self.modelFontInfo.font = font
            ## FIXME: This is slow and wrong for icon view,
            ##        view.reset() does it faster. However, it works
            ##        for table view, see below.
            #for topleft, bottomright in \
            #           modelIterateChildrenRect(self, maxDepth=1):
            #   self.dataChanged.emit(topleft, bottomright)
            return
        self.beginResetModel()
        self._setFont(font, fontInfo, **kwargs)
        self.endResetModel()

    def setMultiFont(self, *args, **kwargs):
        """setMultiFont(current, selected=())

        Set the fonts in the character map. Pass the current
        font as FontSymbolModel() object, and the selected fonts
        as a list of FontSymbolModel() objects.

        Lacks nochange argument, use updateQtFonts().

        FontGrid.setFont() removes the current font from the selected
        fonts, so weird things will happen if you call this directly and
        you do not remove it (particularly, you'd get one more column).
        """
        self.beginResetModel()
        self._setMultiFont(*args, **kwargs)
        self.endResetModel()

    def updateQtFonts(self, currentFont: QtGui.QFont,
                            selectedFonts: IterableOf[QtGui.QFont]=()):
        """Update the qt fonts in the model. Equivalent to nochange=True
        of setFont(). The selected fonts need to provide in the same order
        as the setMultiFont() selected fonts were provided (unless you want
        to do something really weird)."""
        self.modelFontInfo.font = currentFont
        for mfi, font in zip(self.selected, selectedFonts):
            mfi.font = font

        for topleft, bottomright in \
                    modelIterateChildrenRect(self, maxDepth=1):
            self.dataChanged.emit(topleft, bottomright)

    def parent(self, index):
        if not index.isValid():
            return QtCore.QModelIndex()
        return QtCore.QModelIndex()

    def index(self, row, column, parent=QtCore.QModelIndex()):
        if parent.isValid():
            return QtCore.QModelIndex()

        try:
            char = self.characters[row]
        except IndexError:
            return QtCore.QModelIndex()
        else:
            return self.createIndex(row, column, None)

    def hasChildren(self, parent=QtCore.QModelIndex()):
        if not parent.isValid():
            return True
        return False

    def columnCount(self, parent=QtCore.QModelIndex()):
        if parent.isValid():
            # FIXME: If this returns 0, all children disappear. WHY?
            # Documentation says it *must* return 0.
            return 1
        return 1 + len(self.selected)

    def rowCount(self, parent=QtCore.QModelIndex()):
        if parent.column() > 0:
            return 0
        if not parent.isValid():
            return len(self.characters)
        return 0

    def data(self, index, role=Qt.DisplayRole):

        if not index.isValid():
            return None

        char = self.characters[index.row()]

        if role == Qt.DisplayRole:
            if self.charDb.combining(char):
                return '\u25CC' + chr(char)
            else:
                if self.ownCharacters:
                    info = None
                else:
                    info = self.charDb.info_by_code.get(char)
                if info is not None and info.category == 'Cc':
                    if info.display:
                        return chr(info.display)
                    else:
                        return '\u2388'
                else:
                    return chr(char)
        elif role == Qt.EditRole:
            return chr(char)
        elif role == CellHeaderRole:
            charsetCodeInfo = self.charsetCodeInfo
            if charsetCodeInfo is not None:
                result = charsetCodeInfo.get(char)
                if result is None:
                    return '--'
                return result.display
            return "%X" % (char, )
        elif role == SampleRendererRole:
            col = index.column()
            if col == 0:
                return self.sampleRenderers.get(char)
            else:
                return self.selected[col - 1].sampleRenderers.get(char)
        elif role == Qt.FontRole:
            col = index.column()
            if col == 0:
                return self.font
            else:
                return self.selected[col - 1].font
        elif role == Qt.ToolTipRole:
            text = []
            self.fillCharacterInfoTexts(char, text, text, includePreview=True)
            return '<br>'.join(text)
        elif role == CharCodeRole:
            return char
        elif role == Qt.ForegroundRole:
            if self.ownCharacters:
                info = None
            else:
                info = self.charDb.info_by_code.get(char)
            palette = QtWidgets.QApplication.palette()
            if info is not None and info.category == 'Cc':
                return palette.brush(palette.Disabled, palette.Text)
            else:
                return palette.brush(palette.Text)
        elif role == ModelFontInfoRole:
            col = index.column()
            if col == 0:
                return self.modelFontInfo
            else:
                return self.selected[col - 1]

        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                if section == 0:
                    item = self.fontInfo
                else:
                    item = self.selected[section - 1].fontInfo
                if self.singleFamily:
                    return item.translate('style', textlang())
                elif self.singleStyle:
                    return item.translate('family', textlang())
                else:
                    return item.translate('fullname', textlang())

        elif orientation == Qt.Vertical:
            if role == Qt.DisplayRole:
                char = self.characters[section]
                return "%04X" % (char, )

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemFlag()
        return (super(FontSymbolModel, self).flags(index)
                    | Qt.ItemIsEditable | Qt.ItemIsSelectable
                    | Qt.ItemIsDragEnabled)

    def mimeData(self, indexes):
        if not indexes:
            return None

        chars = ''.join(index.data(Qt.EditRole) or ''
                        for index in indexes)
        markup = '<span style="%s">%s</span>' % (qFontToCss(self.font),
                                                 htesc(chars))

        mime = QtCore.QMimeData()
        mime.setText(chars)
        mime.setHtml(markup)
        return mime

    def mimeTypes(self):
        return ['text/html', 'text/plain']

    def dropMimeData(self, data, action, row, column, parent):
        return False

    def supportedDragActions(self):
        return Qt.CopyAction

    def supportedDropActions(self):
        return Qt.CopyAction

    def fillCharacterInfoTexts(self, char: int,
                               infoTextLines: MutableSequenceOf[str]=None,
                               altInfoTextLines: MutableSequenceOf[str]=None,
                               includePreview: bool=False,
                               modelFontInfo: SymbolModelFontInfo=None):
        """Fill the provided sequences with HTML information about the character.

        The Unicode character information goes in the first sequence.
        Any information from alternative registries goes in the second sequence.

        If you provide includePreview=True, the character will also be previewed.
        That mode is used for the tooltip, as the character bars usually have their
        own character preview.
        """

        mfi = modelFontInfo or self

        options = self.renderingChoice.options

        text = infoTextLines
        if text is None:
            text = []

        textAlt = altInfoTextLines
        if textAlt is None:
            textAlt = []

        if self.charDb is None:
            text.append(H_('Unicode database not loaded'))
            return

        info = self.charDb.get(char, autofill=True)
        if info is None:
            text.append(H_('Unknown character'))
            return

        plane = self.charDb.get_plane(char)
        script = self.charDb.find_script_name(self.langDb, char)
        symbol = chr(char)

        display = symbol
        if info.display:
            display = chr(info.display)

        if includePreview:
            text.append('<span style="font-size: %dpt; %s">%s</span>'
                            % (self.renderingChoice.charBoxSize,
                               htesc(qFontToCss(mfi.font)),
                                htesc(chr(char))))

        text.extend([
                htesc(info.name or info.old_name or ''),
                htesc(_('Hex code', ': 0x%x') % (char, )),
                htesc(_('Character', ': %s') % (display, )),
                '%s: %s %s' % (H_('Category'),
                               getIconHtml(info.category_icon,
                                           options.infoIconSize),
                               info.category_name(translate=_)),
        ])

        if info.block:
            text.append(htesc(_('Block', ': %s') % (info.block.name, )))
        text.append(htesc(_('Script', ': %s') % (script, )))
        text.append('%s: %s' % (htesc(_('Plane %d') % (plane.number, )),
                                H_(plane.description)))

        for alias in info.aliases:
            text.append('= ' + htesc(alias))
        for alias in info.formalaliases:
            text.append(htesc(charinfo.formal_alias + ' ' + alias))
        #for crossref in info.crossrefs:
        #    text.append(htesc(charinfo.cross_reference + ' ' + crossref))
        for variation in info.variations:
            text.append(htesc(_('Variation', ': ') + variation))

        if mfi.fontCmap:
            fontCharName = mfi.fontCmap.get(char)
            text.append(htesc(_("Font's character name", ": %s")
                                        % (fontCharName, )))

        altinfos = self.charDb.get_secondary(char)
        if altinfos:
            textAlt.append('')
        for altinfo in altinfos:
            textAlt.append(H_(altinfo.registry.name))
            textAlt.append(htesc(altinfo.name or altinfo.old_name or ''))
            if altinfo.block:
                textAlt.append(htesc(_('Block', ': %s')
                                        % (altinfo.block.name, )))
            altscript = altinfo.registry.find_script_name(self.langDb, char,
                                                          default=None)
            if altscript:
                textAlt.append(htesc(_('Script', ': %s') % (altscript, )))


class GridCellDelegate(QtWidgets.QStyledItemDelegate):

    """A delegate to paint a cell in the character map. This adds
    a header with the hex code of the character, usually with the unicode
    codepoint.

    This header is filled from CellHeaderRole, which may contain:
        - hex code in another encoding, for legacy encodings
        - font name, if the grid is used for displaying font samples
        - morse code or needle telegraph code of the letter

    You can provide the view to get style information from it.

    You can control the grid sizing with square (default True)
    and abbreviateHeader (default False). When square is True,
    effort will be made to make the cells square unless a very wide
    character comes along.  If abbreviateHeader=True is passed,
    and the header is too long, it will be cut off.

    The sizeThreshold is the threshold above which the header is no longer
    displayed.

    You can specify which model roles to use to get the headerRole and
    the sampleRole, they default to CellHeaderRole, and Qt.DisplayRole.
    If you replace the FontSymbolModel with a FontListModel, you'd obviously
    need to change the headers.
    """


    def __init__(self, view: QtWidgets.QAbstractItemView=None, *args,
                       fontDb: QtGui.QFontDatabase=None,
                       square: bool=True, abbreviateHeader: bool=False,
                       headerRole: int=CellHeaderRole,
                       sampleRole: int=Qt.DisplayRole,
                       sampleRendererRole: int=SampleRendererRole,
                       sizeThreshold: numbers.Real=ID_THRESH,
                       **kwargs):
        super(GridCellDelegate, self).__init__(*args, **kwargs)
        self.view = view
        if fontDb is None:
            fontDb = QtGui.QFontDatabase()
        self.fontDb = fontDb

        self.square = square
        self.abbreviateHeader = abbreviateHeader
        self.sampleRole = sampleRole
        self.sizeThreshold = sizeThreshold

        self.headerRole = headerRole
        self.sampleRole = sampleRole
        self.sampleRendererRole = sampleRendererRole

    def sizeHint(self, option, index):

        font = index.data(Qt.FontRole)
        if font is None:
            return super().sizeHint(option, index)

        pointSize = font.pointSize()
        if pointSize >= self.sizeThreshold:

            metrics = QtGui.QFontMetrics(font)
            sample = index.data(self.sampleRole)
            #result = metrics.size(Qt.TextSingleLine, sample)
            result = metrics.size(Qt.TextSingleLine, sample).expandedTo(
                         metrics.boundingRect(sample).size())

            defaultFont = QtWidgets.QApplication.font()
            headerFont = self.fontDb.font(defaultFont.family(),
                                          defaultFont.styleName(),
                                          max(pointSize / 6, 8))
            headerMetrics = QtGui.QFontMetrics(headerFont)
            header = index.data(self.headerRole)
            headerSize = headerMetrics.size(Qt.TextSingleLine, header)

            width = result.width() * 1.1
            if self.square:
                width = max(width, result.height() * 1.1)
            if not self.abbreviateHeader:
                width = max(width, headerSize.width() * 1.1)

            height = result.height() * 1.1 + headerSize.height()

            return QtCore.QSize(width, height)

        result = super().sizeHint(option, index)

        if not self.square:
            return result

        return QtCore.QSize(max(result.width(), result.height()),
                            result.height())

    def paint(self, painter, option, index):
        self.initStyleOption(option, index)

        font = option.font
        if not font:
            return super(GridCellDelegate, self).paint(painter, option, index)

        if self.view is None:
            style = QtWidgets.QApplication.style()
        else:
            style = self.view.style()
        metrics = option.fontMetrics
        palette = option.palette

        darkBackground = hasDarkBackground(palette)

        painter.save()

        rect = style.subElementRect(style.SE_ItemViewItemText, option)
        #rect = option.rect

        pointSize = font.pointSize()
        if pointSize >= self.sizeThreshold:

            # Calculate the headers
            defaultFont = QtWidgets.QApplication.font()
            headerFont = self.fontDb.font(defaultFont.family(),
                                          defaultFont.styleName(),
                                          max(pointSize / 6, 8))
            headerMetrics = QtGui.QFontMetrics(headerFont)

            enabled = False
            if option.state & style.State_Enabled:
                enabled = True

            if option.state & style.State_Selected:
                samplePaintRole = palette.HighlightedText
                headerPaintRole = palette.HighlightedText
            else:
                samplePaintRole = palette.Text
                headerPaintRole = palette.Shadow

            header = index.data(self.headerRole)
            headerSize = headerMetrics.size(Qt.TextSingleLine, header)

            sample = index.data(self.sampleRole)
            sampleRenderer = None

            if self.sampleRendererRole is not None:
                sampleRenderer = index.data(self.sampleRendererRole)

            # Calculate the rectangles
            charOption = type(option)(option)
            charOption.rect = option.rect.adjusted(0, headerSize.height(), 0, 0)
            headerOption = type(option)(option)
            headerOption.rect = option.rect.adjusted(
                                    0, 0, 0, 0 - option.rect.height()
                                               + headerSize.height())

            if headerOption.rect.width() < headerSize.width():
                header = headerMetrics.elidedText(header, Qt.ElideRight,
                                                  headerOption.rect.width())

            # Draw
            if option.state & style.State_Selected:
                painter.fillRect(option.rect, palette.highlight())
            painter.setPen(palette.color(palette.Mid))
            painter.drawRect(option.rect)

            #style.drawControl(style.CE_ItemViewItem, charOption,
            #                  painter, self.view)
            painter.setFont(option.font)
            if sampleRenderer:

                sampleTextRect = style.itemTextRect(metrics, charOption.rect,
                                            Qt.AlignCenter, enabled,
                                            sample)
                sampleRenderer.render(painter, sampleTextRect,
                                      palette, darkBackground)

                #style.drawItemPixmap(painter,
                #                     style.itemPixmapRect(charOption.rect,
                #                                          Qt.AlignCenter,
                #                                          pixmap),
                #                     Qt.AlignCenter, pixmap)
            else:
                style.drawItemText(painter, #charOption.rect,
                                   style.itemTextRect(metrics, charOption.rect,
                                                      Qt.AlignCenter, enabled,
                                                      sample),
                                   Qt.AlignCenter, palette, enabled, sample,
                                   samplePaintRole)
            painter.setFont(headerFont)
            style.drawItemText(painter, #headerOption.rect,
                               style.itemTextRect(headerMetrics,
                                                  headerOption.rect,
                                                  Qt.AlignRight, enabled,
                                                  header),
                               Qt.AlignRight, palette, enabled, header,
                               headerPaintRole)

        else:
            painter.setPen(palette.color(palette.Mid))
            painter.drawRect(option.rect)
            style.drawControl(style.CE_ItemViewItem, option, painter, self.view)

        painter.restore()


class CharFilterModel(QtModelProxies.QSortFilterProxyModel):

    """A filter for character models. User is resposible for setting
    the filterText, filterBlock and filterScript properties and
    calling invalidateFilter()."""

    filterText = None
    filterBlock = None
    filterScript = None
    displayLittle = False

    def __init__(self, charDb: CharacterDatabase=None,
                       langDb: LanguageDatabase=None,
                       parent: QtCore.QObject=None):
        super(CharFilterModel, self).__init__(parent)
        if charDb is None:
            charDb = CharacterDatabase.getInstance()
        if langDb is None:
            langDb = LanguageDatabase.getInstance()
        self.charDb = charDb
        self.langDb = langDb

    def filterAcceptsRow(self, source_row, source_parent):

        if self.filterText:
            # FIXME: Slow-ish for CJK fonts?
            text = self.filterText.strip().casefold()
            char = self.sourceModel().characters[source_row]
            symbol = chr(char)
            if symbol.casefold() == text:
                return True

            if len(text) == 1:
                return False

            try:
                code = int(text, 16)
            except ValueError:
                pass
            else:
                if char == code:
                    return True

            if self.charDb is None:
                return False

            infos = self.charDb.getall(char)
            if not infos:
                return False
            for info in infos:
                if text in (info.name or '').casefold():
                    return True
                if any(text in (alias or '').casefold()
                       for alias in info.aliases):
                    return True
            return False

        # These are taken into account only if text isn't entered
        if self.filterBlock is not None:
            char = self.sourceModel().characters[source_row]
            if not (self.filterBlock.start <= char <= self.filterBlock.end):
                return False

        if self.filterScript is not None:
            if self.charDb is None:
                return self.filterScript == 'Unknown'

            char = self.sourceModel().characters[source_row]
            script = self.charDb.find_script_name(self.langDb, char)
            if script != self.filterScript:
                return False

        if (self.displayLittle and self.filterBlock is None and
            self.filterScript is None):
                char = self.sourceModel().characters[source_row]
                return char < PREVIEW_MAX_CODEPOINT

        return True


class FontGridToolbox(Toolbox):

    """Actions for the font grid, e.g. for its context menu."""

    def __init__(self, grid: 'FontGrid'):
        self.grid = grid
        self._properties = None
        super(FontGridToolbox, self).__init__()

    @property
    def actionParent(self) -> QtCore.QObject:
        return self.grid

    def actionDefinitions(self) -> Iterator:
        yield self.action(self.copyFormatted, 
                          _("Copy formatted characters"), 
                          icon='edit-copy', shortcut='Ctrl+C')
        yield self.copyCharacter, _("Copy plain characters")
        yield self.separator()
        yield self.copyCode, _("Copy decimal code")
        yield partial(self.copyCode, 'hex'), _("Copy hex code")
        yield partial(self.copyCode, 'xml'), _("Copy XML code")
        yield partial(self.copyCode, 'utf8'), _("Copy UTF-8 byte codes")
        yield self.separator()
        yield self.copyName, _("Copy name")
        yield self.copyImage, _("Copy image")

        yield self.separator()
        yield self.action(self.openProperties, _('Character properties...'),
                          icon='document-properties')

    @Slot()
    def copyCharacter(self):
        """Copy the slected character as a single string in plain text"""
        selected = self.grid.view.selectionModel().selectedIndexes()
        if not selected:
            return
        chars = ''.join(index.data(Qt.EditRole) or ''
                        for index in selected)
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(chars)

    @Slot()
    def copyFormatted(self):
        """Copy the selected characters as a single formatted string.
        This gets it from model.mimeData()."""
        selected = self.grid.view.selectionModel().selectedIndexes()
        if not selected:
            return
        model = selected[0].model()
        mime = model.mimeData(selected)
        if mime is None:
            return

        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setMimeData(mime)

    @Slot()
    def copyCode(self, which: str='integer'):
        """Copy the code, either as integer, hex, xml, or utf8"""
        current = self.grid.view.selectionModel().currentIndex()
        if not current.isValid():
            return
        charCode = current.data(CharCodeRole)
        if which == 'integer':
            charCode = str(charCode)
        elif which == 'hex':
            charCode = hex(charCode)
        elif which == 'xml':
            charCode = '&#%d;' % (charCode, )
        elif which == 'utf8':
            charCode = ' '.join(map(hex, chr(charCode).encode('utf8')))

        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(charCode)

    @Slot()
    def copyName(self):
        """Copy the name of the character."""
        current = self.grid.view.selectionModel().currentIndex()
        if not current.isValid():
            return
        charCode = current.data(CharCodeRole)
        info = self.grid.charDb.get(charCode)
        if info is None:
            name = ''
        else:
            name = info.name or info.old_name or ''
        if not name:
            for altinfo in self.grid.charDb.get_secondary(charCode):
                if altinfo.name or altinfo.old_name:
                    name = altinfo.name or altinfo.old_name
                    break

        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(name)

    @Slot()
    def copyImage(self):
        """Copy the character as a 192pt image."""
        current = self.grid.view.selectionModel().currentIndex()
        if not current.isValid():
            return
        font = current.data(Qt.FontRole)
        font = QtGui.QFont(font)
        font.setPointSize(192)
        metrics = QtGui.QFontMetrics(font)
        renderer = current.data(SampleRendererRole)
        char = current.data(Qt.DisplayRole)

        #size = metrics.size(Qt.TextSingleLine, char)
        size = metrics.boundingRect(char).size()
        image = QtGui.QImage(size, QtGui.QImage.Format_ARGB32_Premultiplied)
        image.fill(QtGui.QColor(0, 0, 0, 0))
        painter = QtGui.QPainter()
        painter.begin(image)

        if renderer:
            renderer.render(painter, image.rect())
        else:
            painter.setFont(font)
            painter.drawText(image.rect(), Qt.AlignCenter, char)
        painter.end()

        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setImage(image)

    @Slot()
    def copyScalable(self):
        """Copy scalable, as an SVG."""
        # FIXME: This doesn't work with ColorLayerFontSample, as
        # the glyphs aren't in the cmap, and thus cannot be added to the
        # SVG.

        current = self.grid.view.selectionModel().currentIndex()
        if not current.isValid():
            return
        font = current.data(Qt.FontRole)
        font = QtGui.QFont(font)
        font.setPointSize(192)
        metrics = QtGui.QFontMetrics(font)
        renderer = current.data(SampleRendererRole)
        char = current.data(Qt.DisplayRole)

        #size = metrics.size(Qt.TextSingleLine, char)
        size = metrics.boundingRect(char).size()
        rect = QtCore.QRect(QtCore.QPoint(0, 0), size)
        #image = QtGui.QImage(size, QtGui.QImage.Format_ARGB32_Premultiplied)
        #image.fill(QtGui.QColor(0, 0, 0, 0))
        buf = QtCore.QBuffer()
        image = QtSvg.QSvgGenerator()
        image.setOutputDevice(buf)
        image.setViewBox(rect)
        painter = QtGui.QPainter()
        painter.begin(image)
        painter.setRenderHint(painter.Antialiasing)

        if renderer:
            renderer.render(painter, rect)
        else:
            painter.setFont(font)
            painter.drawText(rect, Qt.AlignCenter, char)
        painter.end()

        mime = QtCore.QMimeData()
        mime.setData("image/svg+xml", buf.buffer())

        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setMimeData(mime)

    @Slot()
    def openProperties(self):
        """Open a dialog with the properties of the character."""
        current = self.grid.view.selectionModel().currentIndex()
        if not current.isValid():
            return

        if self._properties is not None:
            self._properties.hide()
            self._propertiesClosed()

        charCode = current.data(CharCodeRole)
        fontItem = self.grid.fontItem

        self._properties = CharInfoWidget(fontItem, charCode,
                                          charDb=self.grid.charDb,
                                          fontDb=self.grid.fontDb,
                                          langDb=self.grid.langDb)
        self._properties.closed.connect(self._propertiesClosed)
        self._properties.show()

    @Slot()
    def _propertiesClosed(self):
        """Triggered when the properties are closed."""
        if self._properties is not None:
            self._properties.deleteLater()
            self._properties = None


class GridOptionsToolbox(Toolbox):

    """Actions for the font grid options, e.g. for a toolbar in
    the grid, specifying the size of the view."""

    def __init__(self, grid):
        self.grid = grid
        self.renderingChoice = grid.renderingChoice

        super().__init__()

    def actionDefinitions(self) -> Iterator:

        with self.group('size'):
            yield self.action(self.smaller, _('Smaller'),
                              icon='format-font-size-less')
            yield self.action(self.larger, _('Larger'),
                              icon='format-font-size-more')

    def _jump_to_size(self, offset: int):
        """Increate the size with the given offset of jumps. This
        iterates over the .fontsizes property of the grid."""
        sizes = self.grid.fontsizes
        size = self.renderingChoice.gridSize

        if len(sizes) < 2:
            return

        i = min(range(len(sizes)), key=lambda i: abs(sizes[i] - size))
        try:
            size = sizes[i + offset]
        except IndexError:
            return

        self.renderingChoice.setGridSize(size)

    def larger(self):
        """Make the font larger, by picking the next grid.fontsizes
        size."""
        self._jump_to_size(+1)

    def smaller(self):
        """Make the font larger, by picking the previous grid.fontsizes
        size."""
        self._jump_to_size(-1)


class FontGrid(QtWidgets.QWidget):

    """A character map widget."""

    def __init__(self, charDb: CharacterDatabase=None,
                       fontDb: QtGui.QFontDatabase=None,
                       langDb: LanguageDatabase=None,
                       renderingChoice: FontRenderingChoice=None,
                       options: Options=None,
                       histories: datastore.Histories=None,
                       parent: QtWidgets.QWidget=None):
        super(FontGrid, self).__init__(parent)
        self.charDb = charDb

        if options is None:
            options = Options.getInstance()

        if fontDb is None:
            fontDb = QtGui.QFontDatabase()
        if renderingChoice is None:
            renderingChoice = FontRenderingChoice()
        if charDb is None:
            charDb = CharacterDatabase.getInstance()
        if langDb is None:
            langDb = LanguageDatabase.getInstance()
        if histories is None:
            histories = datastore.Histories.getInstance()

        self.fontDb = fontDb
        self.langDb = langDb
        self.histories = histories
        self.renderingChoice = renderingChoice
        self.fontsizes = [12]
        self.fontItem = None
        self.selectedFontItems = []

        self.charBoxes = []
        self.charInfos = []
        self.charInfosAlt = []

        self.currentModelFontInfo = None
        self.currentChar = None

        searchBar = QtWidgets.QToolBar()
        searchBar.setIconSize(options.toolbarIconSize)
        options.toolbarIconSizeChanged.connect(searchBar.setIconSize)

        self.searchText = searchText = QuickSearch(
                history=histories.get_history('fontgrid-characters'),
                placeholderText=_('Search characters...'))
        searchText.searchTriggered.connect(self._filterTextChanged)

        self.blockCombo = blockCombo = CustomComboBox()
        blockCombo.setNarrow()
        blockCombo.currentIndexChanged.connect(self._blockChanged)
        self.blockByIndex = {}

        self.scriptCombo = scriptCombo = CustomComboBox()
        scriptCombo.currentIndexChanged.connect(self._scriptChanged)
        scriptCombo.setNarrow()
        self.scriptByIndex = {}

        self.renderingChoice.gridChanged.connect(self._renderingChanged)

        searchBar.addWidget(searchText)
        searchBar.addWidget(blockCombo)
        searchBar.addWidget(scriptCombo)

        self.model = model = FontSymbolModel(charDb=charDb, langDb=langDb,
                                             renderingChoice=renderingChoice)

        self.internalCharBar = charBar = self.createCharBar()
        self.externalCharBar = None

        self.filterModel = filterModel = CharFilterModel(charDb)
        filterModel.setSourceModel(model)

        self.view = view = self._createGridView()

        model.gridView = view
        model.charBox = self.charBoxes[0]

        self._updateUnicodeSelectors()

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(searchBar)
        layout.addWidget(view, 1)
        layout.addWidget(self.internalCharBar)

        self.setLayout(layout)

        self.toolbox = FontGridToolbox(self)
        self.toolbox.addContextMenu(self.view)
        self.toolbox.disableActions()

        self.optionsToolbox = GridOptionsToolbox(self)
        self.optionsToolbox.populateToolbar(searchBar)

        self.characterCount = QtWidgets.QLabel()
        self.characterCount.setTextFormat(Qt.PlainText)

        self.filterModel.modelReset.connect(self._updateCharacterCount)
        self.filterModel.rowsInserted.connect(self._updateCharacterCount)
        self.filterModel.rowsRemoved.connect(self._updateCharacterCount)

        searchBar.addSeparator()

        self.charBarAction = searchBar.addAction(getIcon('show-character'),
                                                 _('Character information'))
        self.charBarAction.setCheckable(True)
        self.charBarAction.setChecked(True)
        self.charBarAction.toggled.connect(self.toggleCharBar)

        self._updateCharacterCount()

    def _replaceView(self, view: QtWidgets.QAbstractItemView,
                           isListView: bool=True):
        """Replace the view when switching from list view (one-font grid) to
        a table view (multi-font table). Don't forget what view we're using.
        """

        oldView = self.view
        self.model.gridView = view if isListView else None
        self.layout().replaceWidget(oldView, view)
        oldView.deleteLater()
        self.view = view
        self.toolbox.addContextMenu(self.view)

    def _createGridView(self) -> QtWidgets.QListView:
        """Create a new listview font grid view for a single font."""

        view = QtWidgets.QListView()

        view.setViewMode(QtWidgets.QListView.IconMode)
        view.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        view.setMovement(QtWidgets.QListView.Snap)
        view.setFlow(QtWidgets.QListView.LeftToRight)
        view.setWrapping(True)
        view.setLayoutMode(QtWidgets.QListView.Batched)
        #view.setGridSize(QtCore.QSize(40, 40))
        view.setItemDelegateForColumn(0, GridCellDelegate(view,
                                                          fontDb=self.fontDb))
        view.setResizeMode(QtWidgets.QListView.Adjust)
        view.setDragEnabled(True)
        view.setDragDropMode(view.DragOnly)
        view.setSelectionMode(view.ExtendedSelection)

        self.transposedModel = None
        view.setModel(self.filterModel)
        view.selectionModel().currentChanged.connect(self._currentChanged)

        return view

    def _createTableView(self) -> QtWidgets.QTableView:
        """Create a new table view for multiple fonts."""

        view = QtWidgets.QTableView()
        view.setItemDelegate(GridCellDelegate(view, fontDb=self.fontDb))
        view.setDragEnabled(True)
        view.setDragDropMode(view.DragOnly)
        view.setSelectionMode(view.ExtendedSelection)

        #self.transposedModel = transposedModel = TransposedModel()
        #transposedModel.setSourceModel(self.filterModel)
        #view.setModel(transposedModel)

        view.setModel(self.filterModel)
        view.selectionModel().currentChanged.connect(self._currentChanged)

        view.setHorizontalHeader(RotatedHeaderView(Qt.Horizontal, angle=0))

        setResizeMode(view.verticalHeader(),
                      QtWidgets.QHeaderView.ResizeToContents)
        setResizeMode(view.horizontalHeader(),
                      QtWidgets.QHeaderView.ResizeToContents)
        return view

    @Slot(bool)
    def toggleCharBar(self, checked: bool=True):
        """Show or hide the internal character bar, depending on the
        first argument."""
        if checked:
            self.internalCharBar.show()
        else:
            self.internalCharBar.hide()

    def saveDockInnerState(self) -> Mapping:
        """Provide options save state for when used as dock"""
        return {'char-bar-visible': not self.internalCharBar.isHidden()}

    def restoreDockInnerState(self, state: Mapping, strict: bool=False):
        """Restore the save state from the options."""
        if 'char-bar-visible' not in state:
            return
        self.charBarAction.setChecked(state['char-bar-visible'])

    def getCharBar(self) -> QtWidgets.QWidget:
        """Get the external character bar widget, for use as a
        dock in the main window."""
        if self.externalCharBar is None:
            self.externalCharBar = self.createCharBar()
        return self.externalCharBar

    def createCharBar(self) -> QtWidgets.QWidget:
        """Create a new character bar controlled by this font grid.
        You typically use a single external character bar, and
        the internal one, so you do not need to create additional ones.
        """

        layout = QtWidgets.QGridLayout()

        charBox = QtWidgets.QLabel()
        charBox.setTextFormat(Qt.PlainText)
        #charBox.setReadOnly(True)

        charInfo = QtWidgets.QLabel()
        charInfo.setTextFormat(Qt.RichText)

        charInfoAlt = QtWidgets.QLabel()
        charInfoAlt.setTextFormat(Qt.RichText)

        selectedChars = QtWidgets.QTextEdit()
        selectedChars.setSizePolicy(QtWidgets.QSizePolicy.Ignored,
                                    QtWidgets.QSizePolicy.Ignored)
        metrics = QtGui.QFontMetrics(QtGui.QFont())
        selectedChars.setMinimumSize(QtCore.QSize(
                metrics.width('This is it.'),
                metrics.height()))
        selectedChars.setWordWrapMode(
            QtGui.QTextOption.WrapAtWordBoundaryOrAnywhere)

        charInfoAlt.setSizePolicy(QtWidgets.QSizePolicy.Maximum,
                                  QtWidgets.QSizePolicy.Maximum)

        layout.addWidget(charBox, 0, 0, 2, 1)
        layout.addWidget(charInfo, 0, 1, 2, 1)
        layout.addWidget(charInfoAlt, 0, 2, 1, 2)
        layout.addWidget(selectedChars, 1, 2, 1, 2)

        layout.setRowStretch(1, 1)

        self.charBoxes.append(charBox)
        self.charInfos.append(charInfo)
        self.charInfosAlt.append(charInfoAlt)

        self._updateCharBars()

        widget = QtWidgets.QWidget()
        widget.setLayout(layout)
        return widget

    @Slot()
    def _updateCharacterCount(self, *args):
        """Update the character count."""
        self.characterCount.setText(
            _('{count} characters out of {total}').format(
                    total=self.model.rowCount(),
                    count=self.filterModel.rowCount()))

    def _updateUnicodeSelectors(self):
        """Update the unicode selectors."""
        blockCombo = self.blockCombo
        scriptCombo = self.scriptCombo

        self.scriptByIndex.clear()
        self.blockByIndex.clear()
        blockCombo.clear()
        scriptCombo.clear()

        blockCombo.addItem(_('All'))

        unicodeBlocks = self.charDb.unicode_blocks
        scripts = self.charDb.scriptorder

        curBlock = self.filterModel.filterBlock
        curScript = self.filterModel.filterScript
        curBlockIndex = 0
        curScriptIndex = 0
        if self.model.fontCharBlocks:
            available = self.model.fontCharBlocks
            scriptBlocks = []
            for script in scripts:
                for block in self.charDb.scripts[script]:
                    scriptBlocks.append(block)

            unicodeBlocks = blockmath.overlapping_blocks(unicodeBlocks,
                                                         available, True)
            scriptBlocks = blockmath.overlapping_blocks(scriptBlocks,
                                                        available, True)
            scripts = OrderedSet(s.name for s in scriptBlocks)

        for i, block in enumerate(unicodeBlocks, 1):
            if curBlock is not None and curBlock.name == block.name:
                curBlockIndex = i

            self.blockByIndex[i] = block
            blockCombo.addItem(block.name)

        self.scriptCombo.addItem(_('All'))
        for i, script in enumerate(scripts, 1):
            if curScript is not None and curScript == script:
                curScriptIndex = i
            self.scriptByIndex[i] = script
            scriptCombo.addItem(script)

        changed = False

        blockCombo.setCurrentIndex(curBlockIndex)
        if not curBlockIndex:
            if self.filterModel.filterBlock is not None:
                changed = True
            self.filterModel.filterBlock = None

        blockCombo.setCurrentIndex(curBlockIndex)
        if not curScript:
            if self.filterModel.filterScript is not None:
                changed = True
            self.filterModel.filterScript = None

        ## This is a good idea, but the user will be confused.
        #if (not self.model.ownCharacters or
        #    len(self.model.characters) > BIG_FONT_THRESHOLD):

        if not self.model.ownCharacters and (
            self.model.encoding is None or
            self.model.encoding.unicode):

            if not self.filterModel.displayLittle:
                changed = True
            self.filterModel.displayLittle = True

        else:
            if self.filterModel.displayLittle:
                changed = True
            self.filterModel.displayLittle = False

        if changed:
            self.filterModel.invalidateFilter()

    @Slot(str)
    def _filterTextChanged(self, text: bool):
        """The filter text was changed. Update the filter proxy."""
        self.filterModel.filterText = text.lower()
        self.filterModel.invalidateFilter()

    @Slot(int)
    def _blockChanged(self, index: int):
        """The selected unicode block was changed. Update the filter proxy."""
        if index == 0:
            self.filterModel.filterBlock = None
        else:
            self.filterModel.filterBlock = self.blockByIndex.get(index)
        self.filterModel.invalidateFilter()

    @Slot(int)
    def _scriptChanged(self, index: int):
        """The selected script was changed. Update the filter proxy."""
        if index == 0:
            self.filterModel.filterScript = None
        else:
            self.filterModel.filterScript = self.scriptByIndex.get(index)
        self.filterModel.invalidateFilter()

    @Slot(QtCore.QModelIndex)
    def _currentChanged(self, index: QtCore.QModelIndex):
        """The current index of the model was changed. Notify the gang."""
        if not index.isValid():
            self.toolbox.disableActions()
            return

        self.currentChar = index.data(CharCodeRole)
        self.currentModelFontInfo = index.data(ModelFontInfoRole)

        self._currentCharacterChanged()

    @Slot()
    def _currentCharacterChanged(self):
        """The current character was changed. Notify the gang."""
        char = self.currentChar

        if char is None:
            self.toolbox.disableActions()
            return

        self._updateCharBars()
        self.toolbox.enableActions()

    @Slot()
    def _updateCharBars(self):
        """Update the character bars if the character was changed."""
        char = self.currentChar
        if char is None:
            return

        mfi = self.currentModelFontInfo

        for charBox in self.charBoxes:
            charBox.setText(chr(char))
            if mfi is not None:
                charBox.setFont(mfi.font)

        text = []
        textAlt = []
        self.model.fillCharacterInfoTexts(char, text, textAlt,
                                          includePreview=False,
                                          modelFontInfo=mfi)
        for charInfo in self.charInfos:
            charInfo.setText('<br>'.join(text))
        for charInfoAlt in self.charInfosAlt:
            charInfoAlt.setText('<br>'.join(textAlt))

    def modelFontInfoForItem(self, item: 'typeatlas.fontlist.FontLike',
                                   *args, **kwargs) -> SymbolModelFontInfo:
        """Get the SymbolModelFontInfo instance for a font item."""
        options = self.renderingChoice.options
        return SymbolModelFontInfo(self.fontForItem(item), item, *args,
                                   options=options, **kwargs)

    def fontForItem(self, item: 'typeatlas.fontlist.FontLike',
                          forCharBox: bool=False) -> QtGui.QFont:
        """Get the QFont for a font item. The size depends on forCharBox
        state and the rendering choice."""

        if forCharBox:
            size = self.renderingChoice.charBoxSize
        else:
            size = self.renderingChoice.gridSize
        font = self.fontDb.font(item.family, item.style, size)
        font.setStyleHint(QtGui.QFont.AnyStyle,
                          self.renderingChoice.antialiasStyle)
        font.setHintingPreference(self.renderingChoice.hintingPreference)
        return font

    def _setFontSizesFontItem(self, item: 'typeatlas.fontlist.FontLike',
                              selected: 'SequenceOf[typeatlas.fontlist.FontLike]'=()):
        """Set the font sizes for the selected font items."""
        self.fontsizes = self.fontDb.smoothSizes(item.family, item.style)
        if not self.fontsizes:
            self.fontsizes = self.fontDb.standardSizes()
            if not self.fontsizes:
                self.fontsizes = [12]

    def _setCharBoxesFontItem(self, item: 'typeatlas.fontlist.FontLike',
                              selected: 'SequenceOf[typeatlas.fontlist.FontLike]'=()):
        """Change all the character boxes' fonts."""
        charBoxFont = self.fontForItem(item, forCharBox=True)
        for charBox in self.charBoxes:
            charBox.setFont(charBoxFont)

    @Slot(object)
    @Slot(object, list)
    def setFontItem(self, item: 'typeatlas.fontlist.FontLike',
                          selected: 'SequenceOf[typeatlas.fontlist.FontLike]'=(),
                          nochange: bool=False, *args, **kwargs):
        """Set the current font item. You can also pass selected items if
        multiple fonts are selected.

        If nochange=True is passed, only the fonts are updated, not the
        character lists or the overall list of fonts.

        The current font is removed from the selected fonts.
        """

        if selected:
            return self._setMultiFontItems(item, selected,
                                           nochange=nochange,
                                           *args, **kwargs)

        if self.selectedFontItems:
            self._replaceView(self._createGridView())

        if item is None:
            return

        self.fontItem = item
        if not nochange:
            self.selectedFontItems = []

        self._setFontSizesFontItem(item)
        self._setCharBoxesFontItem(item)
        self.model.setFont(self.fontForItem(item), item, nochange=nochange,
                           *args, **kwargs)

        if not nochange:
            self._updateUnicodeSelectors()
        else:
            self.view.reset()

    @Slot(object, list)
    def _setMultiFontItems(self,
                           current: 'typeatlas.fontlist.FontLike',
                           selected: 'SequenceOf[typeatlas.fontlist.FontLike]'=(),
                           nochange: bool=False, *args, **kwargs):
        """Implementation of setFontItem() for non-empty selected."""

        if current is None and not selected:
            return

        if current is None:
            current = selected[0]
        selected = [s for s in selected if s is not current]

        if not selected:
            self.setFontItem(current, nochange=nochange, *args, **kwargs)
            return

        if nochange:
            self.model.updateQtFonts(self.fontForItem(self.fontItem),
                                     map(self.fontForItem,
                                         self.selectedFontItems))
            return

        if not self.selectedFontItems:
            self._replaceView(self._createTableView(), isListView=False)

        self.fontItem = current
        self.selectedFontItems = selected

        self._setFontSizesFontItem(current, selected)
        self._setCharBoxesFontItem(current, selected)
        self.model.setMultiFont(self.modelFontInfoForItem(current, *args, **kwargs),
                                (self.modelFontInfoForItem(item, *args, **kwargs)
                                 for item in selected))

        self._updateUnicodeSelectors()

    @Slot()
    def _renderingChanged(self):
        """The rendering of fonts was chnaged. Re-render the views."""
        self.setFontItem(self.fontItem, self.selectedFontItems,
                         nochange=True)




# -*- coding: utf-8 -*-
#
#    TypeAtlas Unit Tests for Range Math
#    Copyright (C) 2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#

import os
import nose
from nose.tools import assert_equal, assert_not_equal, assert_raises
from nose.tools import assert_sequence_equal, assert_set_equal
from nose.tools import assert_less_equal, assert_greater_equal
from nose.tools import assert_set_equal, assert_in, assert_not_in
from itertools import chain, product

if os.environ.get('TYPEATLAS_DEBUG_TYPECHECK'):
    from typeguard.importhook import install_import_hook
    install_import_hook('typeatlas')

from typeatlas.rangemath import OrdinalRange, CharacterRange, MultiRange
from typeatlas.rangemath import RangeBase, EMPTY_RANGE
from typeatlas.blockmath import intersection, have_intersection, iterblocks
from typeatlas.blockmath import intersect_many, blockslen
from typeatlas.util import OrderedSet

int_members = {-2000, -1000, 1,2,3,4,202,1001, 1662, 100248, 1000000000000000}

sets = [
    set(),
    {1,2,3,5,6,7,10,12,13},
    {2,3,7,10},
    {8},
    {2,3,7,8,10},
    {2,3,7,8,9},
    OrderedSet([100,14,3,13]),
    {101, 100, 14,12},
    {102, 101, 100, 14,13,12},
]

ordinal_range_pairs = [
    (OrdinalRange(1,5), OrdinalRange(-1,7)),
    (OrdinalRange(1,5), OrdinalRange(0,6)),
    (OrdinalRange(1,5), OrdinalRange(1,7)),
    (OrdinalRange(1,5), OrdinalRange(0,5)),
    (OrdinalRange(1,5), OrdinalRange(0,3)),
    (OrdinalRange(1,5), OrdinalRange(3,7)),
    (OrdinalRange(1,5), OrdinalRange(5,7)),
    (OrdinalRange(1,5), OrdinalRange(6,8)),
    (OrdinalRange(1,5), OrdinalRange(5,6)),
    (OrdinalRange(1,5), OrdinalRange(2,3)),
    (OrdinalRange(1,5), OrdinalRange(2,4)),
    (OrdinalRange(1,5), EMPTY_RANGE),
    (MultiRange([OrdinalRange(1,5), OrdinalRange(7,12)]),
     MultiRange([OrdinalRange(-1,3), OrdinalRange(8,15)])),
    (MultiRange([OrdinalRange(1,5), OrdinalRange(7,12)]),
     MultiRange([OrdinalRange(8,15), OrdinalRange(-1,3)])),
    (MultiRange([OrdinalRange(1,5), OrdinalRange(7,12)]),
     MultiRange([OrdinalRange(3,6), OrdinalRange(8,15)])),
    (MultiRange([OrdinalRange(1,5), OrdinalRange(7,12)]),
     MultiRange([OrdinalRange(-1,3), OrdinalRange(6,11)])),
    (MultiRange([OrdinalRange(1,5), OrdinalRange(7,12)]),
     OrdinalRange(3, 11)),
    (MultiRange([OrdinalRange(1,5), OrdinalRange(7,12)]),
     MultiRange([OrdinalRange(-1,3), OrdinalRange(8,15)])),
    (OrdinalRange.fromstring('5-100 303-581 2547-15824 16821-154775'),
     OrdinalRange.fromstring('80-1555 5666-1577922')),

]

char_members = CharacterRange.fromstring('a-z A-Z 0-9')

character_range_pairs = [
    (CharacterRange('a','z'), CharacterRange('A','Z')),
    (CharacterRange('a','z'), CharacterRange('A','c')),
    (CharacterRange('G','z'), CharacterRange('A','Z')),
    (CharacterRange.fromstring('a-c d-f w-z'),
     CharacterRange.fromstring('b-p q-x')),
]


def all_sets():
    yield from sets

def all_ranges():
    yield from chain.from_iterable(ordinal_range_pairs)
    yield from chain.from_iterable(character_range_pairs)

def all_setlike():
    yield from all_sets()
    yield from all_ranges()

def all_pairs():
    yield from ordinal_range_pairs
    yield from character_range_pairs
    yield from product(map(OrdinalRange.from_iterable, sets),
                       map(OrdinalRange.from_iterable, sets))


class TestRange:

    def test_contains(self):
        int_vals = set(int_members)
        for rg in chain.from_iterable(ordinal_range_pairs):
            if len(rg) < 1000:
                int_vals.update(rg)

        for s in chain(all_sets(),
                       chain.from_iterable(ordinal_range_pairs)):
            if isinstance(s, RangeBase):
                rg = s
                s = set(rg)
            else:
                rg = OrdinalRange.from_iterable(s)

            for val in s:
                assert_in(val, rg)
            for val in int_vals:
                if val in s:
                    continue
                assert_not_in(val, rg)

            if rg:
                assert_in(rg.ranges[0].start, rg)
                assert_in(rg.ranges[-1].end, rg)
                assert_in(rg.ranges[0].start, s)
                assert_in(rg.ranges[-1].end, s)

                assert_not_in(rg.ranges[0].start - 1, rg)
                assert_not_in(rg.ranges[-1].end + 1, rg)
                assert_not_in(rg.ranges[0].start - 1, s)
                assert_not_in(rg.ranges[-1].end + 1, s)

        char_vals = set(char_members)
        for rg in chain.from_iterable(character_range_pairs):
            s = set(rg)

            for val in s:
                assert_in(val, rg)
            for val in char_vals:
                if val in s:
                    continue
                assert_not_in(val, rg)

    def test_successor(self):
        int_vals = set(int_members)

        for s in chain(all_sets(),
                       chain.from_iterable(ordinal_range_pairs)):
            if isinstance(s, RangeBase):
                rg = s
                s = set(rg)
            else:
                rg = OrdinalRange.from_iterable(s)

            if not rg:
                continue

            for val in chain(int_vals, s,
                             [rg.ranges[0].start, rg.ranges[0].end]):
                assert_equal(val + 1, rg.successor(val))
                assert_equal(val - 1, rg.predecessor(val))

        char_vals = set(char_members)
        for rg in chain.from_iterable(character_range_pairs):
            s = set(rg)
            for val in chain(char_vals, s,
                             [rg.ranges[0].start, rg.ranges[0].end]):

                assert_equal(chr(ord(val) + 1), rg.successor(val))
                assert_equal(chr(ord(val) - 1), rg.predecessor(val))

    def test_bool(self):
        for s in chain(all_sets(),
                       chain.from_iterable(ordinal_range_pairs),
                       chain.from_iterable(character_range_pairs)):
            if isinstance(s, RangeBase):
                rg = s
                s = set(rg)
            else:
                rg = OrdinalRange.from_iterable(s)

            assert_equal(bool(s), bool(rg))

    def test_round_trip_sets(self):
        for s in all_sets():
            rg = OrdinalRange.from_iterable(s)
            rebuilt = set(rg)
            assert_equal(set(s), set(rg), [s, rg])
            assert_equal(s, rg, [s, rg])
            assert_equal(s, rebuilt, [s, rg, rebuilt])

    def test_round_trip_ranges(self):
        for rg in all_ranges():
            s = set(rg)

            if not s:
                rebuilt = OrdinalRange.from_iterable(s)
            else:
                rebuilt = rg.range_type.from_iterable(s)
            assert_equal(set(s), set(rg), [s, rg])
            assert_equal(s, rg, [s, rg])
            assert_equal(s, rebuilt, [s, rg])

    def test_round_trip_string(self):
        for rg in all_ranges():
            s = rg.tostring()
            if not rg:
                rebuilt = OrdinalRange.fromstring(s)
            else:
                rebuilt = rg.range_type.fromstring(s)
            assert_equal(rebuilt, rg, [rg, rg.tostring()])

    def test_round_trip_fontconfig(self):
        for rg in all_ranges():
            if rg and (rg.value_type is str or rg.ranges[0].start < 0):
                continue
            s = rg.to_fontconfig()
            if not rg:
                rebuilt = OrdinalRange.from_fontconfig(s)
            else:
                rebuilt = rg.range_type.from_fontconfig(s)
            assert_equal(rebuilt, rg, [rg, s, rg.tostring()])

    def test_round_trip_blocks(self):
        for rg in all_ranges():
            if rg and rg.value_type is str:
                continue

            s = rg.to_blocks()
            if not rg:
                rebuilt = OrdinalRange.from_blocks(s)
            else:
                rebuilt = rg.range_type.from_blocks(s)
            assert_equal(rebuilt, rg, [rg, rg.tostring()])

    def test_round_trip_plain_blocks(self):
        for rg in all_ranges():
            if rg and rg.value_type is str:
                continue

            s = rg.to_plain_blocks()
            if not rg:
                rebuilt = OrdinalRange.from_blocks(s)
            else:
                rebuilt = rg.range_type.from_blocks(s)
            assert_equal(rebuilt, rg, [rg, rg.tostring()])

    def test_eq_true(self):
        for rg in all_ranges():
            assert_equal(set(rg), set(rg), rg)
            assert_equal(rg, rg, rg)

    def test_eq_others(self):
        for left, right in all_pairs():
            if set(left) == set(right):
                assert_equal(left, right, [left, right])
                assert_equal(right, left, [left, right])
            else:
                assert_not_equal(left, right, [left, right])
                assert_not_equal(right, left, [left, right])

    def test_len(self):
        for rg in all_ranges():
            s = set(rg)
            assert_less_equal(len(s), len(rg), [rg, s])

        for left, right in all_pairs():
            if len(set(left)) == len(set(right)):
                assert_equal(len(left), len(right), [left, right])
            else:
                assert_not_equal(len(left), (right), [left, right])

    def test_lessequal_true(self):
        for rg in all_ranges():
            assert_less_equal(set(rg), set(rg), rg)
            assert_less_equal(set(rg), rg, rg)
            assert_less_equal(rg, set(rg), rg)
            assert_less_equal(rg, rg, rg)

    def test_lessequal_others(self):
        for left, right in all_pairs():
            if set(left) <= set(right):
                assert_less_equal(left, right, [left, right])
                assert_less_equal(set(left), right, [left, right])
                assert_less_equal(left, set(right), [left, right])
            else:
                assert not left <= right, [left, right]
                assert not left <= set(right), [left, right]
                assert not set(left) <= right, [left, right]

            if set(right) <= set(left):
                assert_less_equal(right, left, [right, left])
                assert_less_equal(set(right), left, [right, left])
                assert_less_equal(right, set(left), [right, left])
            else:
                assert not right <= left, [right, left]
                assert not right <= set(left), [right, left]
                assert not set(right) <= left, [right, left]

    def test_intersection(self):
        for left, right in all_pairs():
            assert_equal(set(left) & set(right),
                         set(left & right),
                         [left, right, left & right])
            assert_equal(set(left) & set(right),
                         set(right & left),
                         [left, right, left & right])

            assert_equal(set(left) & set(right),
                         set(left) & right,
                         [left, right, left & right])
            assert_equal(set(left) & set(right),
                         left & set(right),
                         [left, right, left & right])

    def test_disjoint(self):
        for left, right in all_pairs():
            assert_equal(set(left).isdisjoint(set(right)),
                         left.isdisjoint(right),
                         [left, right, left & right])
            assert_equal(set(left).isdisjoint(set(right)),
                         right.isdisjoint(left),
                         [left, right, left & right])

            assert_equal(set(left).isdisjoint(set(right)),
                         set(left).isdisjoint(right),
                         [left, right, left & right])
            assert_equal(set(left).isdisjoint(set(right)),
                         left.isdisjoint(set(right)),
                         [left, right, left & right])

    def test_union(self):
        for left, right in all_pairs():
            assert_equal(set(left) | set(right),
                         set(left | right),
                         [left, right, left | right])
            assert_equal(set(left) | set(right),
                         set(right | left),
                         [left, right, left | right])

            assert_equal(set(left) | set(right),
                         set(left) | right,
                         [left, right, left | right])
            assert_equal(set(left) | set(right),
                         left | set(right),
                         [left, right, left | right])

    def test_sub(self):
        for left, right in all_pairs():
            assert_equal(set(left) - set(right),
                         set(left - right),
                         [left, right, left - right])
            assert_equal(set(right) - set(left),
                         set(right - left),
                         [left, right, right - left])

            assert_equal(set(left) - set(right),
                         set(left) - right,
                         [left, right, left - right])
            assert_equal(set(left) - set(right),
                         left - set(right),
                         [left, right, left - right])

            assert_equal(set(right) - set(left),
                         set(right) - left,
                         [left, right, right - left])
            assert_equal(set(right) - set(left),
                         right - set(left),
                         [left, right, right - left])

    def test_block_intersection(self):
        for left, right in all_pairs():
            if left and left.value_type is str:
                continue
            if right and right.value_type is str:
                continue
            assert_equal(set(iterblocks(
                                intersect_many(left.to_blocks(),
                                               right.to_blocks()))),
                         left & right,
                         [left, right, left & right])

            assert_equal(set(iterblocks(
                                intersect_many(right.to_blocks(),
                                               left.to_blocks()))),
                         left & right,
                         [left, right, left & right])

    def test_blocks_len(self):
        for rg in all_ranges():
            if rg and rg.value_type is str:
                continue
            blocks = rg.to_blocks()
            assert_less_equal(len(rg), blockslen(blocks), [rg, blocks])

    def test_plain_block_intersection(self):
        for left, right in all_pairs():
            if left and left.value_type is str:
                continue
            if right and right.value_type is str:
                continue
            assert_equal(set(iterblocks(
                                intersect_many(left.to_plain_blocks(),
                                               right.to_plain_blocks()))),
                         left & right,
                         [left, right, left & right])

            assert_equal(set(iterblocks(
                                intersect_many(right.to_plain_blocks(),
                                               left.to_plain_blocks()))),
                         left & right,
                         [left, right, left & right])

    def test_plain_blocks_len(self):
        for rg in all_ranges():
            if rg and rg.value_type is str:
                continue
            blocks = rg.to_plain_blocks()
            assert_less_equal(len(rg), blockslen(blocks), [rg, blocks])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#    TypeAtlas Setup Script
#    Written in 2021 by Milko Krachounov
#
#    This file is part of TypeAtlas.
#
#    To the extent possible under law, Milko Krachunov has waived all copyright
#    and related or neighboring rights to TypeAtlas Setup Script.
#    This software is distributed without any warranty.
#
#    You should have received a copy of the CC0 legalcode along with this
#    work.  If not, see <http://creativecommons.org/publicdomain/zero/1.0/>.
#

from typeatlas import proginfo
from distutils.core import setup

with open('README.md', 'r', encoding='utf8') as readme:
    long_description = readme.read()

setup(
    name=proginfo.PROGRAM_CANON_NAME,
    version=proginfo.VERSION,
    url='https://gitlab.com/milkok/typeatlas',
    description='TypeAtlas font explorer',
    long_description=long_description,
    author=proginfo.AUTHOR,
    author_email=proginfo.EMAIL,
    license='GPLv3',
    platforms=['GNU/Linux', 'POSIX'],
    packages=['typeatlas', 'typeatlas.cli', 'typeatlas.data',
              'typeatlas.foreign'],
    package_data={
        'typeatlas': [
            'icons/README',
            'icons/*.svgz',
            'icons/dark/*.svgz',
            'icons/16x16/*.svgz',
            'icons/widgets/*.svgz',
            'icons/flags/README',
            'icons/flags/*.svgz',
            'images/README',
            'images/*.svgz',
            'images/*.png',
            'i18n/*.po',
        ],
        'typeatlas.data': [
            '*.tsv',
        ]
    },
    data_files = [
        ('share/applications', [
            'extras/TypeAtlas.desktop',
            'extras/GlyphAtlas.desktop',
        ]),
        ('share/icons', [
            'typeatlas/icons/typeatlas.svgz',
            'typeatlas/icons/glyphatlas.svgz',
        ]),
    ],
    requires=[
        'packaging', 'fonttools', 'PyQt5'
    ],
    scripts=['typeatlas-qt', 'glyphatlas-qt', 'glyphatlas-select-qt', 'typefind'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    project_urls={
        'Bug Tracker': 'https://gitlab.com/milkok/typeatlas/-/issues',
        'Github Mirror': 'https://github.com/milkokr/typeatlas/',
        'Screenshots': 'https://imgur.com/a/uoaN94p',
    },
    python_requires='>=3.5',
)



# -*- coding: utf-8 -*-
#
#    TypeAtlas Qt Compatibility
#    Written in 2018-2021 by Milko Krachounov
#
#    This file is part of TypeAtlas.
#
#    To the extent possible under law, Milko Krachunov has waived all copyright
#    and related or neighboring rights to TypeAtlas Qt Compatibility.
#    This software is distributed without any warranty.
#
#    You should have received a copy of the CC0 legalcode along with this
#    work.  If not, see <http://creativecommons.org/publicdomain/zero/1.0/>.
#

# License note:
#
# This module will load Qt and PyQt or PySide and import their utilities.
# At the time of writing Qt and PySide were licensed under the GNU LGPL,
# and PyQt was under the GNU GPL.

"""Support Qt 4, 5, possibly 6 when someone gives it to me, with PySide
and PyQt5, whichever sucks less.

The Qt modules from the respective PyQt5/PySide2 package are imported
in the module's namespace.

TODO: Use __getattr__ when we drop pre-Python 3.7 if this even works
with our code.
"""


import sys
import os
import os.path
from operator import methodcaller
from typeatlas import proginfo
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

DEFAULT_ORDER = [
    'PyQt5',    # Qt 5 sucks at font rendering
    'PySide2',  # Maybe this will not crash
    'PyQt4',    # Crashes? (TODO: find why it segfaults)
    'PySide',   # PySide doesn't have QRawFont, segfaults
]

QT_PYSIDE_VERSION = {4: 'PySide', 5: 'PySide2'}


configuredStyle = None


def _get_qt(order=DEFAULT_ORDER):
    global configuredStyle
    
    exc = None

    order = list(order)

    for name in order:
        if name in sys.modules:
            order = [name]
            break

    filename = os.path.join(proginfo.CONFIG_DIR, proginfo.MAIN_CONFIG)
    if os.path.exists(filename):
        parser = configparser.RawConfigParser()
        parser.read_file(open(filename))
        if parser.has_option('Runtime', 'qt_version'):
            ver = parser.get('Runtime', 'qt_version')
            if ver.isdigit() and ver.isascii():
                preferred = {'PyQt' + ver, QT_PYSIDE_VERSION.get(int(ver))}
                order.sort(key=lambda v: v not in preferred)

        if parser.has_option('Runtime', 'qt_bindings'):
            bindings = parser.get('Runtime', 'qt_bindings')
            if bindings == 'pyqt':
                order.sort(key=lambda v: not v.startswith('PyQt'))
            elif bindings == 'pyside':
                order.sort(key=lambda v: not v.startswith('PySide'))

        if parser.has_option('Runtime', 'qt_style'):
            configuredStyle = parser.get('Runtime', 'qt_style')

    for name in order:
        try:
            if name == 'PyQt5':
                from PyQt5 import QtCore, QtNetwork, QtGui, QtWidgets
                try:
                    from PyQt5 import QtSvg
                except ImportError:
                    QtSvg = None

                return (QtCore, QtNetwork, QtGui, QtWidgets, QtCore, QtSvg,
                        QtCore.pyqtSlot, QtCore.pyqtSignal, 5, False)
            elif name == 'PySide2':
                from PySide2 import QtCore, QtNetwork, QtGui, QtWidgets
                try:
                    from PySide2 import QtSvg
                except ImportError:
                    QtSvg = None

                return (QtCore, QtNetwork, QtGui, QtWidgets, QtCore, QtSvg,
                        QtCore.Slot, QtCore.Signal, 5, True)
            elif name == 'PySide':
                from PySide import QtCore, QtNetwork, QtGui
                try:
                    from PySide import QtSvg
                except ImportError:
                    QtSvg = None

                return (QtCore, QtNetwork, QtGui, QtGui, QtGui, QtSvg,
                        QtCore.Slot, QtCore.Signal, 4, True)
            elif name == 'PyQt4':
                from PyQt4 import QtCore, QtNetwork, QtGui, QtCore
                try:
                    from PyQt4 import QtSvg
                except ImportError:
                    QtSvg = None

                return (QtCore, QtNetwork, QtGui, QtGui, QtGui, QtSvg,
                        QtCore.pyqtSlot,  QtCore.pyqtSignal, 4, False)

        except ImportError as exc:
            print("%s failed to load: %s: %s" % (name,
                                                 type(exc).__name__, exc),
                  file=sys.stderr)

    raise exc
             

(QtCore, QtNetwork, QtGui, QtWidgets, QtModelProxies, QtSvg,
 Slot, Signal, QT_VERSION, usesPySide) = _get_qt()

usesPyQt = not usesPySide
QT_BINDINGS = 'PyQt' if usesPyQt else 'PySide'

Qt = QtCore.Qt

if QtSvg is None:
    del QtSvg


if QT_VERSION == 4:
    def setResizeMode(header, mode):
        return header.setResizeMode(mode)

else:
    def setResizeMode(header, mode):
        return header.setSectionResizeMode(mode)


if usesPySide and QT_VERSION >= 5:
    # Support for bytes in PySide2 is shiboken
    qtGetBytes = methodcaller('data')
else:
    qtGetBytes = bytes


if usesPySide:
    try:
        if QT_VERSION >= 5:
            import shiboken2 as shiboken
        else:
            import shiboken
    except ImportError:
        shiboken = None

    def unwrap(ob):
        return int(shiboken.getCppPointer(ob)[0])


else:
    import sip

    def unwrap(ob):
        return int(sip.unwrapinstance(ob))




add_sample = lambda *a, **kw: None

PC_LOAD_SOURCE = 'PC LOAD LETTER - Wikipedia, 00:22, 31 March 2018 <https://en.wikipedia.org/w/index.php?title=PC_LOAD_LETTER&oldid=833343787>'

add_sample('PC LOAD LETTER',
           language='en', device='LCD', region='US',
           vendor='HP', product='LaserJet 4',
           sources=[PC_LOAD_SOURCE])

add_sample('PC LOAD LEGAL',
           language='en', device='LCD', region='US',
           vendor='HP', product='LaserJet 4',
           sources=[PC_LOAD_SOURCE])

add_sample('PC LOAD A4',
           language='en', device='LCD', region='Europe',
           vendor='HP', product='LaserJet 4',
           sources=[PC_LOAD_SOURCE])


add_sample('Abort, Retry, Fail?',
           alternatives=['Abort, Retry, Ignore, Fail?'],
           context='''
C:\>dir b:

Not ready reading drive A
{text}{cursor}
''',
           language='en', device='textmode',
           vendor='Microsoft', product='MS-DOS',
           sources=['Abort, Retry, Fail? - Wikipedia, 02:28, 20 April 2018 <https://en.wikipedia.org/w/index.php?title=Abort,_Retry,_Fail%3F&oldid=837319661>'])


add_sample('Not a typewriter',
           context="550 {username}: User unknown: {text}",
           product="*nix",
           device="terminal",
           sources=['Not a typewriter - Wikipedia, 08:56, 29 May 2016 <https://en.wikipedia.org/w/index.php?title=Not_a_typewriter&oldid=722633543>'])

add_sample('lp0 on fire',
           context="{logdate} {hostname}: {text}",
           vendor="*nix", 
           device="terminal",
           sources=['lp0 on fire - Wikipedia, 02:41, 7 May 2018 <https://en.wikipedia.org/w/index.php?title=Lp0_on_fire&oldid=840000387>'])


add_sample("Bad command or file name",
           context='''
C:\TYPEATLS>TYPEATLS /TTF C:\FONTS\TIMR.TTF
{text}
''',
           language='en', device='textmode',
           vendor='Microsoft', product='MS-DOS',
           sources=['Bad command or file name - Wikipedia, 18:58, 3 December 2017 <https://en.wikipedia.org/w/index.php?title=Bad_command_or_file_name&oldid=813440129>'])


add_sample('values of β will give rise to dom!',
           vendor="Bell Labs", product='Unix', subject="mv",
           device="terminal", 
           sources=['Odd Comments and Strange Doings in Unix <https://www.bell-labs.com/usr/dmr/www/odd.html>'])




# -*- coding: utf-8 -*-
#
#    TypeAtlas Language Utilities
#    Copyright (C) 2018-2021 Milko Krachounov
#
#    This file is part of TypeAtlas
#
#    TypeAtlas is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    TypeAtlas is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with TypeAtlas.  If not, see <http://www.gnu.org/licenses/>.
#
#                                 ***
#
#    Alternatively, you may use this file (part of TypeAtlas libraries)
#    under the terms of the X11/MIT license as follows:
#
#    Permission is hereby granted, free of charge, to any person
#    obtaining a copy of this software and associated documentation
#    files (the "Software"), to deal in the Software without
#    restriction, including without limitation the rights to use,
#    copy, modify, merge, publish, distribute, sublicense, and/or sell
#    copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following
#    conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#    OTHER DEALINGS IN THE SOFTWARE.
#

"""Language and locale utilities, including:
1. Gettext translate functions _, N_, U_, H_
2. ISO codes for languages, scripts and countries
3. Discovery of font samples for given languages or scripts (actually
   uses the samples module)
"""


import os.path
import errno
import locale
import functools
import gettext
import traceback
import re
import csv
import json
import shutil
import urllib.request
import urllib.parse
from html import escape as htesc
import typeatlas
from typeatlas import proginfo, charinfo
from typeatlas.util import OrderedSet, N_, U_, gettext_tag_regex, errmsgf
from typeatlas.util import generic_type
from typeatlas.samples import has_sample, get_sample, get_samples, ANY, ALL
from typeatlas.samples import available_language_codes, SampleInfo
from typeatlas.samples import LangListType, ScriptListType, ScriptType
from collections import namedtuple
from itertools import cycle, chain
from io import StringIO
try:
    from defusedxml import sax
    from xml.sax.handler import ContentHandler as _SaxContentHandler
except ImportError:
    sax = None
    _SaxContentHandler = object


SetOf = generic_type('Set')
SequenceOf = generic_type('Sequence')
IterableOf = generic_type('Iterable')
IteratorOf = generic_type('Iterator')
TupleOf = generic_type('Tuple')
Literal = generic_type('Literal')
Optional = generic_type('Optional')

PathType = SequenceOf[str]
CharsetType = SetOf[int]


translations = gettext.NullTranslations()
has_translations = False

def _(s: str, suffix: str=None) -> str:
    """Translate the string into current locale. Any alphabetical prefix
    before a pipe is removed (e.g. "abc|")."""

    s = getattr(s, 'gettext_msgid', s)

    if not s:
        return s

    s = translations.gettext(s)
    s = gettext_tag_regex.sub('', s, 1)

    if suffix is not None:
        s += suffix

    return s


def H_(s: str, suffix: str=None) -> str:
    """Translate the string into current locale's language, and escape any
    HTML."""
    return htesc(_(s, suffix))


ISO_3166_1_XML_NAME = 'iso_3166'
ISO_3166_2_XML_NAME = 'iso_3166_2'

ISO_639_XML_NAME = 'iso_639'
ISO_639_3_XML_NAME = 'iso_639_3'
ISO_639_5_XML_NAME = 'iso_639_5'

ISO_15924_XML_NAME = 'iso_15924'

ISO_3166_1_JSON_NAME = 'iso_3166-1'
ISO_639_2_JSON_NAME = 'iso_639-2'
ISO_639_3_JSON_NAME = 'iso_639-3'
ISO_639_5_JSON_NAME = 'iso_639-5'

ISO_15924_JSON_NAME = 'iso_15924'

ISO_3166_TZDATA_FILENAME = 'iso3166.tab'

ISO_CODES_JSON_DOWNLOAD_BASE = 'https://anonscm.debian.org/cgit/pkg-isocodes/iso-codes.git/plain/data/'

ISO_CODES_JSON_CACHE_PATH = os.path.join(proginfo.CACHE_DIR,
                                         'iso-codes', 'json')

ISO_CODES_XML_SEARCH_PATH = ['/usr/share/xml/iso-codes']
ISO_CODES_JSON_SEARCH_PATH = [ISO_CODES_JSON_CACHE_PATH,
                              '/usr/share/iso-codes/json']
TZDATA_SEARCH_PATH = ['/usr/share/zoneinfo']


ENGLISH_NAME = object()
NATIVE_NAME = object()
LOCALE_NAME = object()
AUTO_LOCALE_NAME = object()


# TODO: Move en before extra locales?

LanguageInfo = namedtuple('LanguageInfo',
                            ['iso_names', 'native_names',
                             'code_639_1', 'code_639_2t',
                             'code_639_2b', 'code_639_3', 'code_639_5',
                             'family', 'scope', 'type',
                             'other_names', 'notes'])

language_info_639_1 = {}
language_info_639_2t = {}
language_info_639_2b = {}
language_info_639_3 = {}
language_info_639_5 = {}
language_info = {}

language_info_rfc3066 = {}

def _fill_languages():
    """Initialize language information and fill it in the mappings above."""
    root = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
    langfn = os.path.join(root, 'languages_iso639_1.tsv')
    lines = []

    for line in open(langfn, 'rt', encoding='utf8'):
        if line.startswith(';;'):
            continue
        lines.append(line)

    file = StringIO(''.join(lines))

    for row in csv.DictReader(file, delimiter='\t'):
        info = LanguageInfo(list(map(str.strip,
                                     row['ISO language name'].split(','))),
                            list(map(str.strip,
                                     re.split(r'[/;,(]', row['Native name']))),
                            row['639-1'], row['639-2/T'],
                            row['639-2/B'], row['639-3'], '',
                            row['Language family'], '', '', (),
                            row['Notes'])

        language_info_639_1[info.code_639_1] = info
        language_info_639_2t[info.code_639_2t] = info
        language_info_639_2b[info.code_639_2b] = info
        language_info_639_3[info.code_639_3] = info


    langfn = os.path.join(root, 'languages_iso639_2.tsv')
    lines = []

    for line in open(langfn, 'rt', encoding='utf8'):
        if line.startswith(';;'):
            continue
        lines.append(line)

    file = StringIO(''.join(lines))

    for row in csv.DictReader(file, delimiter='\t'):
        term_code, sep, bib_code = row['639-2'].partition('/')
        term_code = term_code.strip()
        if not sep:
            bib_code = term_code
        else:
            bib_code = bib_code.strip().rstrip('*')

        info = LanguageInfo(list(map(str.strip,
                                     row['ISO language name(s)'].split(';'))),
                            list(map(str.strip,
                                     re.split(r'[/;(]', row['Native name(s)']))),
                            row['639-1'], term_code,
                            bib_code, row['639-3'], row['639-5'],
                            '', row['Scope'], row['Type'],
                            list(map(str.strip,
                                     (row['Other name(s)'] or '').split(';'))),
                            '')

        old_info = None
        if info.code_639_1:
            old_info = language_info_639_1.get(info.code_639_1)

        if old_info is not None:
            language_info_639_1[info.code_639_1] = old_info._replace(
                    scope=info.scope, type=info.type,
                    code_639_5=info.code_639_5,
                    other_names=info.other_names)

        if old_info is None and info.code_639_2t:
            old_info = language_info_639_2t.get(info.code_639_2t)
        if old_info is None and info.code_639_2b:
            old_info = language_info_639_2b.get(info.code_639_2b)
        if old_info is not None:
            info = info._replace(notes=old_info.notes, family=old_info.family)

        if info.code_639_1 and info.code_639_1 not in language_info_639_1:
            language_info_639_1[info.code_639_1] = info

        language_info_639_2t[info.code_639_2t] = info
        language_info_639_2b[info.code_639_2b] = info
        language_info_639_3[info.code_639_3] = info
        language_info_639_5[info.code_639_5] = info

    language_info.update(language_info_639_5)
    language_info.update(language_info_639_3)
    language_info.update(language_info_639_2b)
    language_info.update(language_info_639_2t)
    language_info.update(language_info_639_1)

    # RFC 3066 says to only use the 2-letter code if available,
    # but no issue if we lookup by all the 3-letter codes.
    language_info_rfc3066.update(language_info_639_2t)
    language_info_rfc3066.update(language_info_639_1)


_fill_languages()


locales = set('''
aa_DJ aa_ER aa_ET af_ZA ak_GH am_ET an_ES anp_IN ar_AE ar_BH ar_DZ ar_EG ar_IN 
ar_IQ ar_JO ar_KW ar_LB ar_LY ar_MA ar_OM ar_QA ar_SA ar_SD ar_SS ar_SY ar_TN 
ar_YE as_IN ast_ES ayc_PE az_AZ be_BY bem_ZM ber_DZ ber_MA bg_BG bho_IN bn_BD 
bn_IN bo_CN bo_IN br_FR brx_IN bs_BA byn_ER ca_AD ca_ES ca_FR ca_IT cmn_TW 
crh_UA csb_PL cs_CZ cv_RU cy_GB da_DK de_AT de_BE de_CH de_DE de_LI de_LU doi_IN 
dv_MV dz_BT el_CY el_GR en_AG en_AU en_BW en_CA en_DK en_GB en_HK en_IE en_IN 
en_NG en_NZ en_PH en_SG en_US en_ZA en_ZM en_ZW eo es_AR es_BO es_CL es_CO es_CR 
es_CU es_DO es_EC es_ES es_GT es_HN es_MX es_NI es_PA es_PE es_PR es_PY es_SV 
es_US es_UY es_VE et_EE eu_ES eu_FR fa_IR ff_SN fi_FI fil_PH fo_FO fr_BE fr_CA 
fr_CH fr_FR fr_LU fur_IT fy_DE fy_NL ga_IE gd_GB gez_ER gez_ET gl_ES gu_IN gv_GB 
hak_TW ha_NG he_IL hi_IN hne_IN hr_HR hsb_DE ht_HT hu_HU hy_AM ia ia_FR id_ID 
ig_NG ik_CA is_IS it_CH it_IT iu_CA iw_IL ja_JP ka_GE kk_KZ kl_GL km_KH kn_IN 
kok_IN ko_KR ks_IN ku_TR kw_GB ky_KG lb_LU lg_UG li_BE lij_IT li_NL lo_LA lt_LT 
lv_LV lzh_TW mag_IN mai_IN mg_MG mhr_RU mi_NZ mk_MK ml_IN mni_IN mn_MN mr_IN 
ms_MY mt_MT my_MM nan_TW nb_NO nds_DE nds_NL ne_NP nhn_MX niu_NU niu_NZ nl_AW 
nl_BE nl_NL nn_NO nr_ZA nso_ZA oc_FR om_ET om_KE or_IN os_RU pa_IN pap_AN pap_AW 
pap_CW pa_PK pl_PL ps_AF pt_BR pt_PT quz_PE ro_RO ru_RU ru_UA rw_RW sa_IN sat_IN 
sc_IT sd_IN se_NO shs_CA sid_ET si_LK sk_SK sl_SI so_DJ so_ET so_KE so_SO sq_AL 
sq_MK sr_ME sr_RS ss_ZA st_ZA sv_FI sv_SE sw_KE sw_TZ szl_PL ta_IN ta_LK te_IN 
tg_TJ the_NP th_TH ti_ER ti_ET tig_ER tk_TM tl_PH tn_ZA tr_CY tr_TR ts_ZA tt_RU 
ug_CN uk_UA unm_US ur_IN ur_PK uz_UZ ve_ZA vi_VN wa_BE wae_CH wal_ET wo_SN xh_ZA 
yi_US yo_NG yue_HK zh_CN zh_HK zh_SG zh_TW zu_ZA
'''.split())

try:
    locale.setlocale(locale.LC_MESSAGES, '')
except locale.Error as exc:
    errmsgf("Failed to set locale (LC_MESSAGES): %s: %s",
            type(exc).__name__, exc)
    del exc

deflocale, defencoding = locale.getdefaultlocale()
curlocale, curencoding = locale.getlocale()

_messagelang, _messageencoding = locale.getlocale(locale.LC_MESSAGES) or \
                                 (deflocale, defencoding)

def textlang() -> str:
    """Return the current text language"""
    return _messagelang


langorder = OrderedSet()

if curlocale:
    locales.add(curlocale)
    langorder.add(curlocale.lower().replace('_', '-'))
    langorder.add(curlocale.partition('_')[0].lower())
if deflocale:
    locales.add(deflocale)
    langorder.add(deflocale.lower().replace('_', '-'))
    langorder.add(deflocale.partition('_')[0].lower())

varlocales = list(loc.partition('@')[0].partition('.')[0]
                  for variable in ['LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG']
                  for loc in os.environ.get(variable, '').split(':'))


locales.update(varlocales)
langorder.update(loc.lower().replace('_', '-') for loc in varlocales)
langorder.update(loc.partition('_')[0].lower() for loc in varlocales)


def _autofill_extra_locales():
    """On GNU/Linux and other potential Unix-like operating systems,
    do some detection of known extra locales."""
    if os.path.exists('/etc/locale.gen'):
        for line in sorted(open('/etc/locale.gen'),
                        key=lambda line: line.startswith('en')):
            line = line.strip()
            if line.startswith('#') or not line:
                line = line.replace('#', '').strip()
                if line:
                    loc = line.split()[0]
                    if re.match(r'^[a-z]+_[A-Z]+(?:@|$)', loc):
                        locales.add(loc)
                continue
            loc = line.split()[0]
            locales.add(loc)
            langorder.add(loc.lower().replace('_', '-'))
            langorder.add(loc.partition('_')[0].lower())


langorder.add('en-us')
langorder.add('en')
langimportant = OrderedSet(langorder)
_autofill_extra_locales()


countrylangs = {}
langcountries = {}
countryguess = {'ar': 'arableague', 'en': 'gb'}
langcountrycount = {}

def _autofill_language_guesses():
    """Process the locales and try language x locale counterparts and
    try to decide which flag goes to which language. Uh-oh."""
    for loc in locales:
        lang, sep, country = loc.lower().partition('_')
        if not sep:
            continue

        countrylangs.setdefault(country, []).append(lang)
        langcountries.setdefault(lang, []).append(country)

        # We show courage. We assume that if the country and language code match,
        # the flag of the country flag can represent the language.
        if lang == country:
            countryguess[lang] = country

    for lang, countries in langcountries.items():
        # If only one country speaks that language, and the locale list isn't
        # horribly incomplete, then we could use that country's flag. Possibly.
        # The code checking the second condition is being worked on.
        if len(countries) == 1:
            countryguess[lang] = countries[0]


    for country, langs in countrylangs.items():
        if len(langs) == 1:
            langcountrycount[langs[0]] = \
                langcountrycount.setdefault(langs[0], 0) + 1

    for country, langs in countrylangs.items():
        # If only one country exists such that that a certain language is
        # the only language spoken in the country, according to our database,
        # then make a leap of faith and assume that's a good choice.
        if len(langs) == 1:
            if langs[0] in countryguess:
                continue
            if langcountrycount.get(langs[0]) != 1:
                continue
            countryguess[langs[0]] = country

_autofill_language_guesses()


class _LangXmlHandler(_SaxContentHandler):

    """Parser for XML language files on Unix-like OS. Those are
    deprecated, we are asked to use JSON instead. Whatever."""

    def __init__(self, langdb: 'LanguageDatabase'):
        self.langdb = langdb

    def startElement(self, name, attrs):
        method = getattr(self, 'handle_' + name, None)
        if method is not None:
            method(attrs)

    def handle_iso_639_entry(self, attrs):
        name = attrs.get('name')
        if not name:
            return

        names = tuple(s.strip() for s in name.split(';'))

        for code in [attrs.get('iso_639_1_code'),
                     attrs.get('iso_639_2T_code'),
                     attrs.get('iso_639_2B_code')]:
            if code:
                self.langdb.language_names[code] = names

    def handle_iso_639_3_entry(self, attrs):
        name = attrs.get('reference_name')
        if not name:
            return

        names = tuple(s.strip() for s in name.split(';'))

        code = attrs.get('id')
        if code:
            self.langdb.language_names[code] = names

    def handle_iso_3166_entry(self, attrs):
        name = attrs.get('name')
        if not name:
            return

        for code in [attrs.get('alpha_2_code'),
                     attrs.get('alpha_3_code')]:
            if code:
                self.langdb.country_names[code] = name

    def handle_iso_15924_entry(self, attrs):
        name = attrs.get('name')
        if not name:
            return

        alphacode = attrs.get('alpha_4_code')
        if not alphacode:
            return

        self.langdb.script_names[alphacode] = name

        numcode = attrs.get('numeric_code')

        if numcode and numcode.isdigit() and numcode.isascii():
            numcode = int(numcode)
            self.langdb.script_numcode[alphacode] = numcode
            self.langdb.script_by_numcode[numcode] = alphacode


class LanguageDatabase(object):

    instance = None
    
    def __init__(self,
                 iso_codes_json_search_path: PathType=ISO_CODES_JSON_SEARCH_PATH,
                 iso_codes_xml_search_path: PathType=ISO_CODES_XML_SEARCH_PATH,
                 tzdata_search_path: PathType=TZDATA_SEARCH_PATH,
                 preferred_name: str=LOCALE_NAME):

        self.populated = False

        self.iso_codes_json_search_path = list(iso_codes_json_search_path)
        self.iso_codes_xml_search_path = list(iso_codes_xml_search_path)
        self.tzdata_search_path = list(tzdata_search_path)
        self.preferred_name = preferred_name

        self.language_names = {}
        self.language_native_names = {}

        self.country_names = {}

        self.script_names = {}
        self.script_numcode = {}
        self.script_by_numcode = {}

        self.iso_3166_trans = gettext.translation(ISO_3166_1_XML_NAME,
                                                  fallback=True)
        self.iso_639_trans = gettext.translation(ISO_639_XML_NAME,
                                                 fallback=True)
        self.iso_15924_trans = gettext.translation(ISO_15924_XML_NAME,
                                                   fallback=True)

        self.lang_prio = {lang: i for i, lang in enumerate(langorder)}

        for filename in [ISO_639_3_XML_NAME, ISO_639_5_XML_NAME]:
            try:
                self.iso_639_trans.add_fallback(gettext.translation(filename))
            except EnvironmentError as exc:
                if exc.errno != errno.ENOENT:
                    raise

    @classmethod
    def get_instance(cls, populated: bool=False) -> 'LanguageDatabase':
        """Return the singleton instance of the character database, or create
        one if there is not one created yet.

        If populated=True is passed, the instance is required to be
        populated."""
        if cls.instance is None:
            cls.instance = cls()
        if populated:
            cls.instance.ensure_populated()
        return cls.instance

    getInstance = get_instance

    def downloadables(self) -> IteratorOf[TupleOf[str, str]]:
        """Return pairs of URL, destination path for all the files that
        need to be downloaded for this class to function with latest
        ISO codes."""
        if not os.path.exists(ISO_CODES_JSON_CACHE_PATH):
            os.makedirs(ISO_CODES_JSON_CACHE_PATH)
        for filename in [ISO_639_2_JSON_NAME, ISO_639_3_JSON_NAME,
                         ISO_3166_1_JSON_NAME, ISO_15924_JSON_NAME]:
            url = urllib.parse.urljoin(ISO_CODES_JSON_DOWNLOAD_BASE, filename)
            destpath = os.path.join(ISO_CODES_JSON_CACHE_PATH, filename)
            yield url, destpath

    def download(self):
        """Download the latest ISO codes."""
        for url, destpath in self.downloadables():
            with urllib.request.urlopen(url) as src:
                with open(destpath, 'wb') as dst:
                    shutil.copyfileobj(src, dst)

    def ensure_populated(self, download: bool=False):
        """If ISO codes are not loaded and parsed, parse them. Optionally,
        download them if download=True is passed."""
        if not self.populated:
            if download:
                self.download()
            self.populate()

    def populate(self):
        """Parse the ISO codes, whether downloaded of found on system."""
        for lang, info in language_info.items():
            if not lang:
                continue

            lang = lang.lower()

            if info.iso_names and info.iso_names[0]: 
                self.language_names[lang] = info.iso_names
            if info.native_names and info.native_names[0]:
                self.language_native_names[lang] = info.native_names

        for dirname in self.tzdata_search_path:
            filepath = os.path.join(dirname, ISO_3166_TZDATA_FILENAME)
            if not os.access(filepath, os.R_OK):
                continue

            with open(filepath, 'rt', encoding='utf8') as file:
                for line in file:
                    line, sep, removed = line.partition('#')
                    line = line.strip()
                    if not line:
                        continue
                    try:
                        code, name = line.split(None, 1)
                    except ValueError:
                        continue
                    code = code.upper()
                    self.country_names[code] = name
                break

        for dirname in self.iso_codes_xml_search_path:
            for filename in [ISO_639_XML_NAME, ISO_639_3_XML_NAME,
                             ISO_3166_1_XML_NAME, ISO_15924_XML_NAME]:
                filepath = os.path.join(dirname, filename + '.xml')
                if sax is not None and os.access(filepath, os.R_OK):
                    try:
                        sax.parse(filepath, _LangXmlHandler(self))
                    except Exception:
                        traceback.print_exc()

        def update_lang_dict(d, records, name_key='name', split=True):
            for record in records:
                name = record.get(name_key)
                if not name:
                    continue
                if split:
                    name = tuple(s.strip() for s in name.split(';'))
                for key in ['alpha_2', 'alpha_3']:
                    code = record.get(key)
                    if code:
                        d[code] = name

        for dirname in self.iso_codes_json_search_path:
            filepath = os.path.join(dirname, ISO_639_2_JSON_NAME + '.json')
            if os.access(filepath, os.R_OK):
                with open(filepath) as file:
                    data = json.load(file)
                update_lang_dict(self.language_names, data['639-2'])

            filepath = os.path.join(dirname, ISO_639_3_JSON_NAME + '.json')
            if os.access(filepath, os.R_OK):
                with open(filepath) as file:
                    data = json.load(file)
                update_lang_dict(self.language_names, data['639-3'])

            filepath = os.path.join(dirname, ISO_3166_1_JSON_NAME + '.json')
            if os.access(filepath, os.R_OK):
                with open(filepath) as file:
                    data = json.load(file)
                update_lang_dict(self.country_names, data['3166-1'],
                                 split=False)

            filepath = os.path.join(dirname, ISO_15924_JSON_NAME + '.json')
            if os.access(filepath, os.R_OK):
                with open(filepath) as file:
                    data = json.load(file)

                for record in data['15924']:
                    recname = record.get('name')
                    recalpha = record.get('alpha_4')
                    recnum = record.get('numeric')

                    if not recname or not recalpha:
                        continue

                    self.script_names[recalpha] = recname
                    if recnum and (isinstance(recnum, int) or
                                   (recnum.isdigit() and recnum.isascii())):
                        self.script_numcode[recalpha] = recnum
                        self.script_by_numcode[recnum] = recalpha

        self.populated = True

    def language_name(self, code: str, which=None) -> str:
        """Get the English, localised or native name for a language."""
        if which is None:
            which = self.preferred_name

        if which is AUTO_LOCALE_NAME:
            which = LOCALE_NAME if has_translations else ENGLISH_NAME

        if '_' in code:
            code, sep, country = code.partition('_')
        else:
            code, sep, country = code.partition('-')

        if sep:
            return "%s (%s)" % (self.language_name(code, which),
                                self.country_name(country, which))

        code = code.lower()

        if which is NATIVE_NAME:
            result = self.language_native_names.get(code)
            if result is not None:
                return result[0]
        
        iso_name = self.language_names.get(code, (code, ))[0]

        if which is LOCALE_NAME:
            if not iso_name:
                return ''
            return self.iso_639_trans.gettext(iso_name)
                        
        return iso_name

    def country_name(self, code: str, which=None) -> str:
        """Get the English or localised name for a country."""
        if which is None:
            which = self.preferred_name

        if which is AUTO_LOCALE_NAME:
            which = LOCALE_NAME if has_translations else ENGLISH_NAME

        code = code.upper()

        iso_name = self.country_names.get(code, code)

        if which is not ENGLISH_NAME:
            if not iso_name:
                return ''
            return self.iso_3166_trans.gettext(iso_name)

        return iso_name

    def script_name(self, code: str, which=None,
                          chardb: 'typeatlas.charinfo.CharacterDatabase'=None
                          ) -> str:
        """Get the English or localised name for a script code.

        If chardb is passed, you can also pass the unicode name."""
        if which is None:
            which = self.preferred_name

        if code is ANY or code is ALL or code is None:
            return _('script|Unknown')

        if which is AUTO_LOCALE_NAME:
            which = LOCALE_NAME if has_translations else ENGLISH_NAME

        if isinstance(code, int) or (isinstance(code, str) and
                                     code.isdigit() and code.isascii()):
            code = self.script_by_numcode.get(int(code), str(code))

        elif chardb is not None:
            unicode_aliases = chardb.value_aliases['sc']
            if code in unicode_aliases:
                if code == 'Common':
                    return _('script|Common') \
                          if which is ENGLISH_NAME else code

                elif code == 'Unknown':
                    return _('script|Unknown') \
                           if which is ENGLISH_NAME else code

                elif code == 'Inherited':
                    return _('script|Inherited') \
                           if which is ENGLISH_NAME else code

                code = unicode_aliases[code]
            else:
                code = code.replace('_', ' ')

        iso_name = self.script_names.get(code, code)

        if which is not ENGLISH_NAME:
            if not iso_name:
                return ''
            return self.iso_15924_trans.gettext(iso_name)

        return iso_name

    def samples_langs(self, langs: LangListType, alternative: int=0,
                            long: bool=False, charset: SetOf[int]=None,
                            **kwargs) -> IteratorOf[SampleInfo]:
        """Yield language samples for font testing. They can be selected by
        a list of languages (or ANY if unknown), or font-supported
        charset (a set of unicode codepoints).

        If a charset is provided, the samples are filtered to only include
        those fully supported by the font.

        You can also provide all arguments supported by get_samples(),
        such as exact, fallbacks, script, scripts. See the documentation there.

        If multiple samples are supported for the same language, you can
        pass alternative=1, 2, 3, etc. to get an alternative one. This can
        be used to vary the result to e.g. build longer text.
        """
        if langs is ANY:
            langs = available_language_codes(long=long)

        langcodes = self.sort_languages(langs)
        result = get_samples(langcodes, exact=True, fallbacks=False, long=long,
                             alternative=alternative, **kwargs)
        if charset is not None:
            has_braille = charinfo.BRAILLE_RANGE <= charset

            result = (sample for sample in result
                      if all(ord(c) in charset for c in sample.text))

            if has_braille:
                prepend = []
                braille_sample = None
                for lang in chain(langorder):
                    if has_sample(lang, script='Brai', long=long):
                        braille_sample = get_sample(lang, script='Brai',
                                                    long=long,
                                                    alternative=alternative)
                        prepend.append(braille_sample)
                        break

                if braille_sample is not None:
                    #first_sample = next(result, None)
                    #prepend.insert(0, first_sample)
                    #result = chain(iter(prepend), result)
                    result = chain(result, [braille_sample])

        return result

    def samples_scripts(self, scripts: ScriptListType,
                              alternative: int=0, long: bool=False,
                              charset: SetOf[int]=None,
                              **kwargs) -> IteratorOf[SampleInfo]:
        """Like samples_langs(), but select by scripts."""

        # FIXME: Slow, this loops over all languages for the given scripts
        return self.samples_langs(ANY, alternative=alternative,
                                  long=long, charset=charset,
                                  scripts=scripts, **kwargs)

    def samples_font(self, font: 'typeatlas.fontlist.FontLike',
                           long: bool=False,
                           alternative: int=ANY) -> IteratorOf[SampleInfo]:
        """Get suitable samples for a given font."""
        if font.lang_unknown():
            langs = ANY
        else:
            langs = OrderedSet()
            for lang in font.lang:
                langs.add(lang)
                lang, sep, loc = lang.partition('-')
                if sep:
                    langs.add(lang)

        if alternative is ANY:
            if not font.scalable or not font.outline:
                alternative = 2
            elif not font.sfnt:
                alternative = 1
            else:
                alternative = 0

        return self.samples_langs(langs, alternative, long=long,
                                  charset=font.charset)

    def samples_font_intersection(self,
                                  fonts: 'IterableOf[typeatlas.fontlist.FontLike]',
                                  long: bool=False,
                                  alternative: int=ANY) -> IteratorOf[SampleInfo]:
        """Get suitable samples that cover an intersection of fonts."""
        known_langs = False
        common_langs = None
        common_charset = None
        charset_supported = True
        for font in fonts:
            if font.lang_unknown():
                continue
            known_langs = True

            langs = OrderedSet()
            for lang in font.lang:
                langs.add(lang)
                lang, sep, loc = lang.partition('-')
                if sep:
                    langs.add(lang)

            if not langs:
                continue

            if common_langs is None:
                common_langs = langs
            else:
                common_langs &= langs

            if font.charset is None:
                charset_supported = False
            elif common_charset is None:
                common_charset = font.charset
            else:
                common_charset = common_charset & font.charset

        if not common_langs:
            common_langs = []
        if not known_langs:
            common_langs = ANY

        if not charset_supported:
            common_charset = None

        if alternative is ANY:
            if not any(font.scalable or font.outline for font in fonts):
                alternative = 2
            elif not any(font.sfnt for font in fonts):
                alternative = 1
            else:
                alternative = 0

        return self.samples_langs(common_langs, alternative, long=long,
                                  charset=common_charset)

    def sample_texts_font(self, font: 'typeatlas.fontlist.FontLike'
                         ) -> IteratorOf[str]:
        """Get just the sample texts for a font."""
        return (sample.text for sample in self.samples_font(font))

    def sample_texts_font_intersection(
                        self, fonts: 'IterableOf[typeatlas.fontlist.FontLike]'
                        ) -> IteratorOf[str]:
        """Get just the sample texts usable for an intersection of fonts."""
        return (sample.text
                for sample in self.samples_font_intersection(fonts))

    def sample_language(self, lang: str, *args, **kwargs) -> Optional[SampleInfo]:
        """Get a sample for a given language."""
        return get_sample(lang, *args, **kwargs)

    def sample_text_language(self, *args, **kwargs) -> Optional[str]:
        """Get the sample text for a given language."""
        sample = self.sample_language(*args, **kwargs)
        if sample is not None:
            return sample.text
        return None

    def sample_text_complex(self, lang: str,
                                  script: ScriptType=ALL,
                                  scripts: ScriptListType=None,
                                  min_len: int=2500, max_len: int=8500,
                                  paragraphs: SequenceOf[int]=[100, 400, 800, 500],
                                  titlesize: int=24, subtitlesize: int=40,
                                  extra: SequenceOf[str]=[],
                                  chardb: 'typeatlas.charinfo.CharacterDatabase'=None
                                  ) -> dict:
        """Generate more complex multi-paragraph sample for a language.

        Return a dictionary containing 'title', 'subtitle' and 'text'
        """
        paragraph_text = [[]]
        seen = set()

        title = ''
        subtitle = ''
        length = 0
        paragraph_length = 0
        use_spaces = False

        samples = self.samples_langs(langs=ANY if lang is ANY else [lang],
                                     script=script, scripts=scripts,
                                     alternative=ALL)
        sample_texts = chain(extra or (),
                             (sample.text for sample in samples))

        for i, sample_text in cycle(enumerate(sample_texts)):
            if ' ' in sample_text:
                use_spaces = True

            if i in seen:
                limit = min_len
            else:
                limit = max_len

            if length > limit:
                if not title:
                    title = sample_text
                    continue
                elif not subtitle:
                    subtitle = sample_text
                    continue
                else:
                    break

            size = paragraphs[(len(paragraph_text) - 1) % len(paragraphs)]
            if paragraph_length > size:
                paragraph_text.append([])
                paragraph_length = 0

            seen.add(i)
            paragraph_text[-1].append(sample_text)
            length += len(sample_text)
            paragraph_length += len(sample_text)

        sep = ' ' if use_spaces else ''
        text = '\n\n'.join(sep.join(paragraph) for paragraph in paragraph_text)
        title = excerpt(title, titlesize, chardb=chardb)
        subtitle = excerpt(subtitle, subtitlesize, chardb=chardb)

        return dict(text=text, title=title, subtitle=subtitle)

    def languages_with_samples(self, languages: IterableOf[str],
                                     sort: bool=True,
                                     *args, **kwargs) -> IteratorOf[str]:
        """Yield languages that have samples.

        Languages are sorted unless sort=False is passed. The remainder
        arguments are passed to LanguageSamples.has_sample(), see the
        documentation there or in get_sample().
        """

        if sort:
            languages = self.sort_languages(languages)
        for lang in languages:
            if has_sample(lang, *args, **kwargs):
                yield lang

    def sort_languages(self, languages: IterableOf[str]) -> SequenceOf[str]:
        """Do a loose sort of the languages by the priority as detected
        on the system."""
        return sorted(languages,
                      key=lambda lang: (self.lang_prio.get(lang, 99999999),
                                                           lang))

    def guess_country_flag(self, lang: str) -> Optional[str]:
        """Guess the flag for the given languages. Languages don't have flags,
        so expect surprising results."""
        lang = lang.lower()
        if '-' in lang:
            return lang.rpartition('-')[2]
        return countryguess.get(lang)


def filter_samples(samples: IterableOf[SampleInfo],
                   one_per_script: bool=False,
                   keep_langs: SetOf[str]=langimportant):

    """Filter samples. If one_per_script=True is passed,
    only one sample per script is displayed. You can define
    which languages to keep even if we already got a sample for this
    script. By default, those would be the important languages detected
    on the system."""
    scripts = set()

    for sample in samples:
        if one_per_script:
            if sample.script in scripts and sample.code not in keep_langs:
                continue
            scripts.add(sample.script)
        yield sample


def excerpt(phrase: str, approxlen: int=64, *,
            spaced: bool=None, japanese: bool=None,
            chardb: 'typeatlas.charinfo.CharacterDatabase'=None) -> str:
    """Return a short excerpt from the given phrase, with the specified
    approximate length. It tries to detect spaces when splitting, and
    treat Japanese text sanely, though it will probably fail.

    You can control that behaviour with spaced=True/False, japanese=True/False.
    """
    try:
        import unicodedata
    except ImportError:
        unicodedata = None

    scripts = {}

    if spaced is None:
        spaced = ' ' in phrase

    elif japanese is None and chardb is not None:
        if (phrase and
            chardb.find_script_code(ord(phrase[0]))
                    in ['Hira', 'Kana', 'Hani']):

                scripts = {c: chardb.find_script_code(ord(c)) for c in phrase}
                scriptset = set(scripts.values())
                japanese = 'Hira' in scriptset or 'Kana' in scriptset

    elif japanese and chardb is not None:
        scripts = {c: chardb.find_script_code(ord(c)) for c in phrase}

    result = []
    length = 0
    lastchar = ''

    for i, char in enumerate(phrase):
        code = ord(char)

        if chardb is None:
            length += 1
        else:
            length += chardb.guess_width(code, 0)

        if length >= approxlen:
            if spaced:
                if char == ' ':
                    break

            elif japanese:
                script = scripts.get(char)
                if script == 'Hani':
                    if scripts.get(lastchar) in ['Hira', 'Kana']:
                        break
                elif script not in ['Hira', 'Kana']:
                    break

            else:
                if length >= approxlen:
                    result.append(char)
                    break

        result.append(char)
        lastchar = char

    return ''.join(result)


i18n_root = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'i18n')
mocache_root = os.path.join(proginfo.CACHE_DIR, 'i18n')


def install_locales(languages: IterableOf[str]=None):
    """Set up locales, to be executed at start of TypeAtlas."""
    try:
        locale.setlocale(locale.LC_ALL, '')
    except locale.Error as exc:
        errmsgf("Failed to set locale (LC_ALL): %s: %s", type(exc).__name__, exc)

    if languages is None:
        # FIXME: Use locale through textlang() instead of crawling the variables again?
        for variable in ['LC_ALL', 'LC_MESSAGES', 'LANG', 'LANGUAGE']:
            value = os.environ.get(variable)
            if value:
                languages = OrderedSet(value.split(':'))
                break

    existing = []

    maker = TranslationMaker()
    for lang in languages:
        if maker.exists(lang):
            maker.autocompile(lang)
            existing.append(maker.compiled_path(lang))

    global translations, has_translations

    if not existing:
        translations = gettext.NullTranslations()
        return

    translations = None
    for path in existing:
        if translations is None:
            translations = gettext.GNUTranslations(open(path, 'rb'))
        else:
            translations.add_fallback(gettext.GNUTranslations(open(path, 'rb')))
        has_translations = True


class TranslationMaker(object):

    """A translation loader that can load them directly from source and
    cache the compiled gettext messages in e.g. ~/.cache/typeatlas.
    """

    def autocompile(self, loc: str, _exact_match: bool=False):
        """If there is a source translation for the given locale, but it is
        not compiled, automatically compile it."""
        if not _exact_match:
            loc = locale.normalize(loc).partition('@')[0].partition('.')[0]

        source = os.path.join(i18n_root, loc + '.po')
        compiled = os.path.join(mocache_root, loc + '.mo')

        if not os.path.exists(source):
            if os.path.exists(compiled):
                return
            langcode, sep, loccode = loc.partition('_')
            if sep:
                self.autocompile(langcode, _exact_match=True)
                return
            raise LookupError("no such translation %r" % (loc, ))

        if (os.path.exists(compiled) and
            os.stat(compiled).st_mtime >= os.stat(source).st_mtime):
                return

        try:
            os.makedirs(os.path.dirname(compiled))
        except EnvironmentError as exc:
            if exc.errno != errno.EEXIST:
                raise

        from typeatlas.foreign import msgfmt
        msgfmt.MESSAGES.clear()
        msgfmt.make(source, compiled)

    def exists(self, loc: str, _exact_match: bool=False) -> bool:
        """Return True if the compiled translation for the given language
        exists."""

        if not _exact_match:
            loc = locale.normalize(loc).partition('@')[0].partition('.')[0]

        source = os.path.join(i18n_root, loc + '.po')
        compiled = os.path.join(mocache_root, loc + '.mo')

        if os.path.exists(source):
            return True
        if os.path.exists(compiled):
            return True

        if not _exact_match:
            langcode, sep, loccode = loc.partition('_')
            if sep:
                return self.exists(langcode, _exact_match=True)

        return False

    def compiled_path(self, loc: str, _exact_match: bool=False) -> str:
        """Get the path to the compiled gettext translation for a given
        language / locale. You can call autocompile() before this to
        ensure a cached version exists. Raises LookupError if not found."""
        if not _exact_match:
            loc = locale.normalize(loc).partition('@')[0].partition('.')[0]

        if self.exists(loc, _exact_match=True):
            return os.path.join(mocache_root, loc + '.mo')

        if not _exact_match:
            langcode, sep, loccode = loc.partition('_')
            if sep:
                return self.compiled_path(langcode, _exact_match=True)

        raise LookupError("no compiled translation found for %r" % (loc, ))

